import unittest

from test.util import check_error_message, check_startswith_error_message
from weaviate.gql.filter import (
    NearText,
    NearVector,
    NearObject,
    NearImage,
    MediaType,
    NearVideo,
    NearAudio,
    NearDepth,
    NearThermal,
    NearIMU,
    Where,
    Ask,
    WHERE_OPERATORS,
    VALUE_TYPES,
)


def helper_get_test_filter(filter_type, value):
    return {"path": ["name"], "operator": "Equal", filter_type: value}


class TestNearText(unittest.TestCase):
    def move_x_test_case(self, move: str):
        """
        Test the "moveTo" or the "moveAwayFrom" clause.

        Parameters
        ----------
        move : str
            The "moveTo" or the "moveAwayFrom" clause name.
        """

        type_error_msg = (
            lambda dt: f"'moveXXX' key-value is expected to be of type <class 'dict'> but is {dt}!"
        )
        concepts_objects_error_msg = "The 'move' clause should contain `concepts` OR/AND `objects`!"
        objects_type_error_msg = (
            lambda dt: f"'objects' key-value is expected to be of type (<class 'list'>, <class 'dict'>) but is {dt}!"
        )
        object_value_error_msg = (
            "Each object from the `move` clause should have ONLY `id` OR `beacon`!"
        )
        concept_value_error_msg = lambda dt: (
            f"'concepts' key-value is expected to be of type (<class 'list'>, <class 'str'>) but is {dt}!"
        )
        force_error_msg = "'move' clause needs to state a 'force'"
        force_type_error_msg = lambda dt: (
            f"'force' key-value is expected to be of type <class 'float'> but is {dt}!"
        )

        with self.assertRaises(TypeError) as error:
            NearText({"concepts": "Some_concept", move: "0.5"})
        check_error_message(self, error, type_error_msg(str))

        with self.assertRaises(ValueError) as error:
            NearText({"concepts": "Some_concept", move: {}})
        check_error_message(self, error, concepts_objects_error_msg)

        with self.assertRaises(TypeError) as error:
            NearText({"concepts": "Some_concept", move: {"concepts": set("something")}})
        check_error_message(self, error, concept_value_error_msg(set))

        with self.assertRaises(ValueError) as error:
            NearText(
                {
                    "concepts": "Some_concept",
                    move: {
                        "concepts": "something",
                    },
                }
            )
        check_error_message(self, error, force_error_msg)

        with self.assertRaises(TypeError) as error:
            NearText(
                {
                    "concepts": "Some_concept",
                    move: {
                        "objects": 1234,
                    },
                }
            )
        check_error_message(self, error, objects_type_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearText(
                {
                    "concepts": "Some_concept",
                    move: {
                        "objects": {},
                    },
                }
            )
        check_error_message(self, error, object_value_error_msg)

        with self.assertRaises(ValueError) as error:
            NearText(
                {
                    "concepts": "Some_concept",
                    move: {
                        "objects": {"id": 1, "beacon": 2},
                    },
                }
            )
        check_error_message(self, error, object_value_error_msg)

        with self.assertRaises(ValueError) as error:
            NearText(
                {
                    "concepts": "Some_concept",
                    move: {
                        "objects": {"test_id": 1},
                    },
                }
            )
        check_error_message(self, error, object_value_error_msg)

        with self.assertRaises(TypeError) as error:
            NearText(
                {
                    "concepts": "Some_concept",
                    move: {"concepts": "something", "objects": [{"id": 1}], "force": True},
                }
            )
        check_error_message(self, error, force_type_error_msg(bool))

    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls
        content_error_msg = f"NearText filter is expected to be type dict but is {list}"
        concept_error_msg = "No concepts in content"
        concept_value_error_msg = lambda actual_type: (
            f"'concepts' key-value is expected to be of type (<class 'list'>, <class 'str'>) but is {actual_type}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )
        autocorrect_error_msg = lambda dtype: (
            f"'autocorrect' key-value is expected to be of type <class 'bool'> but is {dtype}!"
        )

        ## test "concepts"
        with self.assertRaises(TypeError) as error:
            NearText(["concepts", "Some_concept"])
        check_error_message(self, error, content_error_msg)

        with self.assertRaises(ValueError) as error:
            NearText({"INVALID": "Some_concept"})
        check_error_message(self, error, concept_error_msg)

        with self.assertRaises(TypeError) as error:
            NearText({"concepts": set("Some_concept")})
        check_error_message(self, error, concept_value_error_msg(set))

        ## test "certainty"
        with self.assertRaises(TypeError) as error:
            NearText({"concepts": "Some_concept", "certainty": "0.5"})
        check_error_message(self, error, certainty_error_msg(str))

        ## test "certainty"
        with self.assertRaises(TypeError) as error:
            NearText({"concepts": "Some_concept", "autocorrect": [True]})
        check_error_message(self, error, autocorrect_error_msg(list))

        ## test "moveTo"
        self.move_x_test_case("moveTo")
        ## test "moveAwayFrom"
        self.move_x_test_case("moveAwayFrom")

        # test valid calls
        NearText({"concepts": "Some_concept"})
        NearText({"concepts": ["Some_concept", "Some_concept_2"]})
        NearText({"concepts": "Some_concept", "certainty": 0.75})
        NearText({"concepts": "Some_concept", "certainty": 0.75, "autocorrect": True})
        NearText(
            {"concepts": "Some_concept", "moveTo": {"concepts": "moveToConcepts", "force": 0.75}}
        )
        NearText(
            {
                "concepts": "Some_concept",
                "moveAwayFrom": {"concepts": "moveAwayFromConcepts", "force": 0.75},
            }
        )
        NearText(
            {
                "concepts": "Some_concept",
                "certainty": 0.75,
                "moveAwayFrom": {"concepts": "moveAwayFromConcepts", "force": 0.75},
                "moveTo": {"concepts": "moveToConcepts", "force": 0.75},
                "autocorrect": False,
            }
        )

        NearText(
            {
                "concepts": "Some_concept",
                "certainty": 0.75,
                "moveAwayFrom": {"objects": {"id": "test_id"}, "force": 0.75},
                "moveTo": {
                    "concepts": "moveToConcepts",
                    "objects": [{"id": "test_id"}, {"beacon": "Test_beacon"}],
                    "force": 0.75,
                },
                "autocorrect": True,
            }
        )

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_text = NearText({"concepts": "Some_concept"})
        self.assertEqual(str(near_text), 'nearText: {concepts: ["Some_concept"]} ')

        near_text = NearText({"concepts": ["Some_concept", "Some_concept_2"]})
        self.assertEqual(
            str(near_text), 'nearText: {concepts: ["Some_concept", "Some_concept_2"]} '
        )
        near_text = NearText({"concepts": "Some_concept", "certainty": 0.75})
        self.assertEqual(str(near_text), 'nearText: {concepts: ["Some_concept"] certainty: 0.75} ')
        near_text = NearText({"concepts": "Some_concept", "autocorrect": True})
        self.assertEqual(
            str(near_text), 'nearText: {concepts: ["Some_concept"] autocorrect: true} '
        )
        near_text = NearText({"concepts": "Some_concept", "autocorrect": False})
        self.assertEqual(
            str(near_text), 'nearText: {concepts: ["Some_concept"] autocorrect: false} '
        )
        near_text = NearText(
            {"concepts": "Some_concept", "moveTo": {"concepts": "moveToConcepts", "force": 0.75}}
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] moveTo: {force: 0.75 concepts: ["moveToConcepts"]}} ',
        )
        near_text = NearText(
            {
                "concepts": "Some_concept",
                "moveTo": {
                    "concepts": "moveToConcepts",
                    "force": 0.75,
                    "objects": {"id": "SOME_ID"},
                },
            }
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] moveTo: {force: 0.75 concepts: ["moveToConcepts"] objects: [{id: "SOME_ID"} ]}} ',
        )

        near_text = NearText(
            {
                "concepts": "Some_concept",
                "moveTo": {
                    "force": 0.75,
                    "objects": [{"id": "SOME_ID"}, {"beacon": "SOME_BEACON"}],
                },
            }
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] moveTo: {force: 0.75 objects: [{id: "SOME_ID"} {beacon: "SOME_BEACON"} ]}} ',
        )

        near_text = NearText(
            {
                "concepts": "Some_concept",
                "moveAwayFrom": {
                    "concepts": "moveAwayFromConcepts",
                    "force": 0.75,
                    "objects": {"id": "SOME_ID"},
                },
            }
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] moveAwayFrom: {force: 0.75 concepts: ["moveAwayFromConcepts"] objects: [{id: "SOME_ID"} ]}} ',
        )

        near_text = NearText(
            {
                "concepts": "Some_concept",
                "moveAwayFrom": {
                    "force": 0.75,
                    "objects": [{"id": "SOME_ID"}, {"beacon": "SOME_BEACON"}],
                },
            }
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] moveAwayFrom: {force: 0.75 objects: [{id: "SOME_ID"} {beacon: "SOME_BEACON"} ]}} ',
        )

        near_text = NearText(
            {
                "concepts": "Some_concept",
                "moveAwayFrom": {"concepts": "moveAwayFromConcepts", "force": 0.25},
            }
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] moveAwayFrom: {force: 0.25 concepts: ["moveAwayFromConcepts"]}} ',
        )
        near_text = NearText(
            {
                "concepts": "Some_concept",
                "certainty": 0.95,
                "moveAwayFrom": {"concepts": "moveAwayFromConcepts", "force": 0.75},
                "moveTo": {"concepts": "moveToConcepts", "force": 0.25},
                "autocorrect": True,
            }
        )
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["Some_concept"] certainty: 0.95 moveTo: {force: 0.25 concepts: ["moveToConcepts"]} moveAwayFrom: {force: 0.75 concepts: ["moveAwayFromConcepts"]} autocorrect: true} ',
        )

        # test it with references of objects
        concepts = ["con1", "con2"]
        move = {"concepts": "moveToConcepts", "force": 0.75}

        near_text = NearText({"concepts": concepts, "moveTo": move})
        concepts.append("con3")  # should not be appended to the nearText clause
        move["force"] = 2.00  # should not be appended to the nearText clause
        self.assertEqual(
            str(near_text),
            'nearText: {concepts: ["con1", "con2"] moveTo: {force: 0.75 concepts: ["moveToConcepts"]}} ',
        )


class TestNearVector(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # test exceptions
        content_error_msg = "NearVector filter is expected to " f"be type dict but is {list}"
        vector_error_msg = "\"No 'vector' key in `content` argument.\""
        vector_value_error_msg = (
            "The type of the 'vector' argument is not supported!\n"
            "Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` "
            "and `tf.Tensor`"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        ## test "concepts"
        with self.assertRaises(TypeError) as error:
            NearVector(["concepts", "Some_concept"])
        check_error_message(self, error, content_error_msg)

        with self.assertRaises(KeyError) as error:
            NearVector({"INVALID": "Some_concept"})
        check_error_message(self, error, vector_error_msg)

        with self.assertRaises(TypeError) as error:
            NearVector({"vector": set("Some_concept")})
        check_error_message(self, error, vector_value_error_msg)

        ## test "certainty"
        with self.assertRaises(TypeError) as error:
            NearVector({"vector": [1.0, 2.0, 3.0, 4.0], "certainty": "0.5"})
        check_error_message(self, error, certainty_error_msg(str))

        # test valid calls
        NearVector({"vector": [1.0, 2.0, 3.0, 4.0]})
        NearVector({"vector": [1.0, 2.0, 3.0, 4.0], "certainty": 0.75})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_vector = NearVector({"vector": [1.0, 2.0, 3.0, 4.0]})
        self.assertEqual(str(near_vector), "nearVector: {vector: [1.0, 2.0, 3.0, 4.0]} ")
        near_vector = NearVector({"vector": [1.0, 2.0, 3.0, 4.0], "certainty": 0.75})
        self.assertEqual(
            str(near_vector), "nearVector: {vector: [1.0, 2.0, 3.0, 4.0] certainty: 0.75} "
        )


class TestNearObject(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        content_error_msg = lambda dt: f"NearObject filter is expected to be type dict but is {dt}"
        beacon_id_error_msg = "The 'content' argument should contain EITHER `id` OR `beacon`!"
        beacon_id_type_error_msg = lambda what, dt: (
            f"'{what}' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearObject(123, is_server_version_14=False)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearObject({"id": 123, "beacon": 456}, is_server_version_14=False)
        check_error_message(self, error, beacon_id_error_msg)

        with self.assertRaises(TypeError) as error:
            NearObject(
                {
                    "id": 123,
                },
                is_server_version_14=False,
            )
        check_error_message(self, error, beacon_id_type_error_msg("id", int))

        with self.assertRaises(TypeError) as error:
            NearObject(
                {
                    "beacon": {123},
                },
                is_server_version_14=False,
            )
        check_error_message(self, error, beacon_id_type_error_msg("beacon", set))

        with self.assertRaises(TypeError) as error:
            NearObject({"beacon": "test_beacon", "certainty": False}, is_server_version_14=False)
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearObject(
            {
                "id": "test_id",
            },
            is_server_version_14=False,
        )

        NearObject({"beacon": "test_beacon", "certainty": 0.7}, is_server_version_14=False)

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearObject(
            {
                "id": "test_id",
            },
            is_server_version_14=False,
        )
        self.assertEqual(str(near_object), 'nearObject: {id: "test_id"} ')

        near_object = NearObject({"id": "test_id", "certainty": 0.7}, is_server_version_14=False)
        self.assertEqual(str(near_object), 'nearObject: {id: "test_id" certainty: 0.7} ')

        near_object = NearObject(
            {
                "beacon": "test_beacon",
            },
            is_server_version_14=False,
        )
        self.assertEqual(str(near_object), 'nearObject: {beacon: "test_beacon"} ')

        near_object = NearObject(
            {"beacon": "test_beacon", "certainty": 0.0}, is_server_version_14=False
        )
        self.assertEqual(str(near_object), 'nearObject: {beacon: "test_beacon" certainty: 0.0} ')


class TestNearImage(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        content_error_msg = lambda dt: f"NearImage filter is expected to be type dict but is {dt}"
        image_key_error_msg = '"content" is missing the mandatory key "image"!'
        image_value_error_msg = (
            lambda dt: f"'image' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearImage(123)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearImage({"id": "image_path.png", "certainty": 456})
        check_error_message(self, error, image_key_error_msg)

        with self.assertRaises(TypeError) as error:
            NearImage({"image": True})
        check_error_message(self, error, image_value_error_msg(bool))

        with self.assertRaises(TypeError) as error:
            NearImage({"image": b"True"})
        check_error_message(self, error, image_value_error_msg(bytes))

        with self.assertRaises(TypeError) as error:
            NearImage({"image": "the_encoded_image", "certainty": False})
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearImage(
            {
                "image": "test_image",
            }
        )

        NearImage({"image": "test_image_2", "certainty": 0.7})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearImage(
            {
                "image": "test_image",
            }
        )
        self.assertEqual(str(near_object), 'nearImage: {image: "test_image"} ')

        near_object = NearImage({"image": "test_image", "certainty": 0.7})
        self.assertEqual(str(near_object), 'nearImage: {image: "test_image" certainty: 0.7} ')


class TestWhere(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # test exceptions
        content_error_msg = lambda dt: f"Where filter is expected to be type dict but is {dt}"
        content_key_error_msg = "Filter is missing required fields `path` or `operands`. Given: "
        path_key_error = "Filter is missing required field `operator`. Given: "
        dtype_no_value_error_msg = "'value<TYPE>' field is either missing or incorrect: "
        dtype_multiple_value_error_msg = "Multiple fields 'value<TYPE>' are not supported: "
        operator_error_msg = (
            lambda op: f"Operator {op} is not allowed. Allowed operators are: {', '.join(WHERE_OPERATORS)}"
        )
        geo_operator_value_type_mismatch_msg = (
            lambda op, vt: f"Operator {op} requires a value of type valueGeoRange. Given value type: {vt}"
        )

        with self.assertRaises(TypeError) as error:
            Where(None)
        check_error_message(self, error, content_error_msg(type(None)))

        with self.assertRaises(TypeError) as error:
            Where("filter")
        check_error_message(self, error, content_error_msg(str))

        with self.assertRaises(ValueError) as error:
            Where({})
        check_startswith_error_message(self, error, content_key_error_msg)

        with self.assertRaises(ValueError) as error:
            Where({"path": "some_path"})
        check_startswith_error_message(self, error, path_key_error)

        with self.assertRaises(ValueError) as error:
            Where({"path": "some_path", "operator": "Like"})
        check_startswith_error_message(self, error, dtype_no_value_error_msg)

        with self.assertRaises(ValueError) as error:
            Where({"path": "some_path", "operator": "Like", "valueBoolean": True, "valueInt": 1})
        check_startswith_error_message(self, error, dtype_multiple_value_error_msg)

        with self.assertRaises(ValueError) as error:
            Where({"operands": "some_path"})
        check_startswith_error_message(self, error, path_key_error)

        with self.assertRaises(TypeError) as error:
            Where({"operands": "some_path", "operator": "Like"})
        check_error_message(self, error, content_error_msg(str))

        with self.assertRaises(TypeError) as error:
            Where({"operands": ["some_path"], "operator": "Like"})
        check_error_message(self, error, content_error_msg(str))

        with self.assertRaises(ValueError) as error:
            Where({"path": "some_path", "operator": "NotValid"})
        check_error_message(self, error, operator_error_msg("NotValid"))

        with self.assertRaises(ValueError) as error:
            Where({"path": "some_path", "operator": "WithinGeoRange", "valueBoolean": True})
        check_error_message(
            self, error, geo_operator_value_type_mismatch_msg("WithinGeoRange", "valueBoolean")
        )

        # test valid calls
        Where({"path": "hasTheOneRing", "operator": "Equal", "valueBoolean": False})
        Where(
            {
                "operands": [
                    {"path": "hasTheOneRing", "operator": "Equal", "valueBoolean": False},
                    {"path": "hasFriend", "operator": "Equal", "valueText": "Samwise Gamgee"},
                ],
                "operator": "And",
            }
        )

    def test___str__(self):
        """
        Test the `__str__` method.
        """
        value_is_not_list_err = (
            lambda v, t: f"Must provide a list when constructing where filter for {t} with {v}"
        )

        test_filter = {"path": ["name"], "operator": "Equal", "valueString": "A"}
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueString: "A"} ', result)

        test_filter = {
            "operator": "Or",
            "operands": [
                {"path": ["name"], "operator": "Equal", "valueString": "Alan Truing"},
                {"path": ["name"], "operator": "Equal", "valueString": "John von Neumann"},
            ],
        }
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {operator: Or operands: [{path: ["name"] operator: Equal valueString: "Alan Truing"}, {path: ["name"] operator: Equal valueString: "John von Neumann"}]} ',
            result,
        )

        # test dataTypes
        test_filter = helper_get_test_filter("valueText", "Test")
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueText: "Test"} ', result)

        test_filter = helper_get_test_filter("valueString", "Test")
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueString: "Test"} ', result)

        test_filter = helper_get_test_filter("valueText", "n\n")
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueText: "n "} ', result)

        test_filter = helper_get_test_filter("valueString", 'what is an "airport"?')
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: Equal valueString: "what is an \\"airport\\"?"} ',
            result,
        )

        # test_filter = helper_get_test_filter("valueString", "this is an escape sequence \a")
        # result = str(Where(test_filter))
        # self.assertEqual(
        #     'where: {path: ["name"] operator: Equal valueString: "this is an escape sequence \\u0007"} ',
        #     result,
        # )

        # test_filter = helper_get_test_filter("valueString", "this is a hex value \u03A9")
        # result = str(Where(test_filter))
        # self.assertEqual(
        #     'where: {path: ["name"] operator: Equal valueString: "this is a hex value \\u03a9"} ',
        #     result,
        # )

        test_filter = helper_get_test_filter("valueText", "what is an 'airport'?")
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: Equal valueText: "what is an \'airport\'?"} ', result
        )

        test_filter = helper_get_test_filter("valueInt", 1)
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueInt: 1} ', result)

        test_filter = helper_get_test_filter("valueNumber", 1.4)
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueNumber: 1.4} ', result)

        test_filter = helper_get_test_filter("valueBoolean", True)
        result = str(Where(test_filter))
        self.assertEqual('where: {path: ["name"] operator: Equal valueBoolean: true} ', result)

        test_filter = helper_get_test_filter("valueDate", "test-2021-02-02")
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: Equal valueDate: "test-2021-02-02"} ', result
        )

        geo_range = {
            "geoCoordinates": {"latitude": 51.51, "longitude": -0.09},
            "distance": {"max": 2000},
        }
        test_filter = helper_get_test_filter("valueGeoRange", geo_range)
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: Equal valueGeoRange: { geoCoordinates: { latitude: 51.51 longitude: -0.09 } distance: { max: 2000 }}} ',
            str(result),
        )

        test_filter = {
            "path": ["name"],
            "operator": "ContainsAny",
            "valueTextArray": ["A", "B\n"],
        }
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: ContainsAny valueText: ["A","B "]} ', str(result)
        )

        test_filter = {
            "path": ["name"],
            "operator": "ContainsAll",
            "valueStringArray": ["A", '"B"'],
        }
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: ContainsAll valueString: ["A","\\"B\\""]} ',
            str(result),
        )

        test_filter = {"path": ["name"], "operator": "ContainsAny", "valueIntArray": [1, 2]}
        result = str(Where(test_filter))
        self.assertEqual(
            'where: {path: ["name"] operator: ContainsAny valueInt: [1, 2]} ', str(result)
        )

        test_filter = {
            "path": ["name"],
            "operator": "ContainsAny",
            "valueStringArray": "A",
        }
        with self.assertRaises(TypeError) as error:
            str(Where(test_filter))
        check_error_message(self, error, value_is_not_list_err("A", "valueStringArray"))

        test_filter = {
            "path": ["name"],
            "operator": "ContainsAll",
            "valueTextArray": "A",
        }
        with self.assertRaises(TypeError) as error:
            str(Where(test_filter))
        check_error_message(self, error, value_is_not_list_err("A", "valueTextArray"))

        test_filter = {
            "path": ["name"],
            "operator": "Equal",
            "valueWrong": "whatever",
        }
        with self.assertRaises(ValueError) as error:
            str(Where(test_filter))
        assert (
            error.exception.args[0]
            == f"'value<TYPE>' field is either missing or incorrect: {test_filter}. Valid values are: {VALUE_TYPES}."
        )


class TestAskFilter(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # test exceptions
        ## error messages
        content_type_msg = lambda dt: f"Ask filter is expected to be type dict but is {dt}"
        question_value_msg = 'Mandatory "question" key not present in the "content"!'
        question_type_msg = (
            lambda dt: f"'question' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_type_msg = (
            lambda dt: f"'certainty' key-value is expected to be of type <class 'float'> but is {dt}!"
        )
        properties_type_msg = (
            lambda dt: f"'properties' key-value is expected to be of type (<class 'list'>, <class 'str'>) but is {dt}!"
        )
        autocorrect_type_msg = (
            lambda dt: f"'autocorrect' key-value is expected to be of type <class 'bool'> but is {dt}!"
        )

        with self.assertRaises(TypeError) as error:
            Ask(None)
        check_error_message(self, error, content_type_msg(type(None)))

        with self.assertRaises(ValueError) as error:
            Ask({"certainty": 0.1})
        check_error_message(self, error, question_value_msg)

        with self.assertRaises(TypeError) as error:
            Ask({"question": ["Who is the president of USA?"]})
        check_error_message(self, error, question_type_msg(list))

        with self.assertRaises(TypeError) as error:
            Ask({"question": "Who is the president of USA?", "certainty": "1.0"})
        check_error_message(self, error, certainty_type_msg(str))

        with self.assertRaises(TypeError) as error:
            Ask({"question": "Who is the president of USA?", "autocorrect": {"True"}})
        check_error_message(self, error, autocorrect_type_msg(set))

        with self.assertRaises(TypeError) as error:
            Ask(
                {
                    "question": "Who is the president of USA?",
                    "certainty": 0.8,
                    "properties": ("prop1", "prop2"),
                }
            )
        check_error_message(self, error, properties_type_msg(tuple))

        # valid calls

        content = {
            "question": "Who is the president of USA?",
        }
        ask = Ask(content=content)
        self.assertEqual(str(ask), f"ask: {{question: \"{content['question']}\"}} ")

        content = {
            "question": "Who is the president of USA?",
            "certainty": 0.8,
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask),
            (
                f"ask: {{question: \"{content['question']}\""
                f' certainty: {content["certainty"]}}} '
            ),
        )

        content = {
            "question": 'Who is the president of "USA"?',
            "certainty": 0.8,
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask),
            (
                f'ask: {{question: "Who is the president of \\"USA\\"?"'
                f' certainty: {content["certainty"]}}} '
            ),
        )

        content = {
            "question": "Who is the president of USA?",
            "certainty": 0.8,
            "properties": "prop1",
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask),
            (
                f"ask: {{question: \"{content['question']}\""
                f' certainty: {content["certainty"]}'
                f' properties: ["prop1"]}} '
            ),
        )

        content = {
            "question": "Who is the president of USA?",
            "certainty": 0.8,
            "properties": ["prop1", "prop2"],
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask),
            (
                f"ask: {{question: \"{content['question']}\""
                f' certainty: {content["certainty"]}'
                f' properties: ["prop1", "prop2"]}} '
            ),
        )

        content = {
            "question": "Who is the president of USA?",
            "certainty": 0.8,
            "properties": ["prop1", "prop2"],
            "autocorrect": True,
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask),
            (
                f"ask: {{question: \"{content['question']}\""
                f' certainty: {content["certainty"]}'
                ' properties: ["prop1", "prop2"] autocorrect: true} '
            ),
        )

        content = {"question": "Who is the president of USA?", "autocorrect": False}
        ask = Ask(content=content)
        self.assertEqual(
            str(ask), (f"ask: {{question: \"{content['question']}\" autocorrect: false}} ")
        )

        content = {
            "question": "Who is the president of USA?",
            "certainty": 0.8,
            "properties": ["prop1", "prop2"],
            "autocorrect": True,
            "rerank": True,
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask),
            (
                f"ask: {{question: \"{content['question']}\""
                f' certainty: {content["certainty"]}'
                ' properties: ["prop1", "prop2"] autocorrect: true'
                " rerank: true} "
            ),
        )

        content = {
            "question": "Who is the president of USA?",
            "rerank": False,
        }
        ask = Ask(content=content)
        self.assertEqual(
            str(ask), (f"ask: {{question: \"{content['question']}\"" " rerank: false} ")
        )


class TestNearVideo(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        media_type = MediaType.VIDEO
        arg_name = media_type.value.capitalize()
        content_error_msg = (
            lambda dt: f"Near{arg_name} filter is expected to be type dict but is {dt}"
        )
        media_key_error_msg = f'"content" is missing the mandatory key "{media_type.value}"!'
        media_value_error_msg = (
            lambda dt: f"'{media_type.value}' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearVideo(123)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearVideo({"id": "video_path.avi", "certainty": 456})
        check_error_message(self, error, media_key_error_msg)

        with self.assertRaises(TypeError) as error:
            NearVideo({"video": True})
        check_error_message(self, error, media_value_error_msg(bool))

        with self.assertRaises(TypeError) as error:
            NearVideo({"video": b"True"})
        check_error_message(self, error, media_value_error_msg(bytes))

        with self.assertRaises(TypeError) as error:
            NearVideo({"video": "the_encoded_video", "certainty": False})
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearVideo(
            {
                "video": "test_video",
            }
        )

        NearVideo({"video": "test_video_2", "certainty": 0.7})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearVideo(
            {
                "video": "test_video",
            }
        )
        self.assertEqual(str(near_object), 'nearVideo: {video: "test_video"} ')

        near_object = NearVideo({"video": "test_video", "certainty": 0.7})
        self.assertEqual(str(near_object), 'nearVideo: {video: "test_video" certainty: 0.7} ')


class TestNearAudio(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        media_type = MediaType.AUDIO
        arg_name = media_type.value.capitalize()
        content_error_msg = (
            lambda dt: f"Near{arg_name} filter is expected to be type dict but is {dt}"
        )
        media_key_error_msg = f'"content" is missing the mandatory key "{media_type.value}"!'
        media_value_error_msg = (
            lambda dt: f"'{media_type.value}' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearAudio(123)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearAudio({"id": "audio_path.wav", "certainty": 456})
        check_error_message(self, error, media_key_error_msg)

        with self.assertRaises(TypeError) as error:
            NearAudio({"audio": True})
        check_error_message(self, error, media_value_error_msg(bool))

        with self.assertRaises(TypeError) as error:
            NearAudio({"audio": b"True"})
        check_error_message(self, error, media_value_error_msg(bytes))

        with self.assertRaises(TypeError) as error:
            NearAudio({"audio": "the_encoded_audio", "certainty": False})
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearAudio(
            {
                "audio": "test_audio",
            }
        )

        NearAudio({"audio": "test_audio_2", "certainty": 0.7})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearAudio(
            {
                "audio": "test_audio",
            }
        )
        self.assertEqual(str(near_object), 'nearAudio: {audio: "test_audio"} ')

        near_object = NearAudio({"audio": "test_audio", "certainty": 0.7})
        self.assertEqual(str(near_object), 'nearAudio: {audio: "test_audio" certainty: 0.7} ')


class TestNearDepth(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        media_type = MediaType.DEPTH
        arg_name = media_type.value.capitalize()
        content_error_msg = (
            lambda dt: f"Near{arg_name} filter is expected to be type dict but is {dt}"
        )
        media_key_error_msg = f'"content" is missing the mandatory key "{media_type.value}"!'
        media_value_error_msg = (
            lambda dt: f"'{media_type.value}' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearDepth(123)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearDepth({"id": "depth_path.png", "certainty": 456})
        check_error_message(self, error, media_key_error_msg)

        with self.assertRaises(TypeError) as error:
            NearDepth({"depth": True})
        check_error_message(self, error, media_value_error_msg(bool))

        with self.assertRaises(TypeError) as error:
            NearDepth({"depth": b"True"})
        check_error_message(self, error, media_value_error_msg(bytes))

        with self.assertRaises(TypeError) as error:
            NearDepth({"depth": "the_encoded_depth", "certainty": False})
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearDepth(
            {
                "depth": "test_depth",
            }
        )

        NearDepth({"depth": "test_depth_2", "certainty": 0.7})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearDepth(
            {
                "depth": "test_depth",
            }
        )
        self.assertEqual(str(near_object), 'nearDepth: {depth: "test_depth"} ')

        near_object = NearDepth({"depth": "test_depth", "certainty": 0.7})
        self.assertEqual(str(near_object), 'nearDepth: {depth: "test_depth" certainty: 0.7} ')


class TestNearThermal(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        media_type = MediaType.THERMAL
        arg_name = media_type.value.capitalize()
        content_error_msg = (
            lambda dt: f"Near{arg_name} filter is expected to be type dict but is {dt}"
        )
        media_key_error_msg = f'"content" is missing the mandatory key "{media_type.value}"!'
        media_value_error_msg = (
            lambda dt: f"'{media_type.value}' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearThermal(123)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearThermal({"id": "thermal_path.png", "certainty": 456})
        check_error_message(self, error, media_key_error_msg)

        with self.assertRaises(TypeError) as error:
            NearThermal({"thermal": True})
        check_error_message(self, error, media_value_error_msg(bool))

        with self.assertRaises(TypeError) as error:
            NearThermal({"thermal": b"True"})
        check_error_message(self, error, media_value_error_msg(bytes))

        with self.assertRaises(TypeError) as error:
            NearThermal({"thermal": "the_encoded_thermal", "certainty": False})
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearThermal(
            {
                "thermal": "test_thermal",
            }
        )

        NearThermal({"thermal": "test_thermal_2", "certainty": 0.7})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearThermal(
            {
                "thermal": "test_thermal",
            }
        )
        self.assertEqual(str(near_object), 'nearThermal: {thermal: "test_thermal"} ')

        near_object = NearThermal({"thermal": "test_thermal", "certainty": 0.7})
        self.assertEqual(str(near_object), 'nearThermal: {thermal: "test_thermal" certainty: 0.7} ')


class TestNearIMU(unittest.TestCase):
    def test___init__(self):
        """
        Test the `__init__` method.
        """

        # invalid calls

        ## error messages
        media_type = MediaType.IMU
        arg_name = media_type.value.upper()
        content_error_msg = (
            lambda dt: f"Near{arg_name} filter is expected to be type dict but is {dt}"
        )
        media_key_error_msg = f'"content" is missing the mandatory key "{media_type.value}"!'
        media_value_error_msg = (
            lambda dt: f"'{media_type.value}' key-value is expected to be of type <class 'str'> but is {dt}!"
        )
        certainty_error_msg = lambda dtype: (
            f"'certainty' key-value is expected to be of type <class 'float'> but is {dtype}!"
        )

        with self.assertRaises(TypeError) as error:
            NearIMU(123)
        check_error_message(self, error, content_error_msg(int))

        with self.assertRaises(ValueError) as error:
            NearIMU({"id": "imu_path.txt", "certainty": 456})
        check_error_message(self, error, media_key_error_msg)

        with self.assertRaises(TypeError) as error:
            NearIMU({"imu": True})
        check_error_message(self, error, media_value_error_msg(bool))

        with self.assertRaises(TypeError) as error:
            NearIMU({"imu": b"True"})
        check_error_message(self, error, media_value_error_msg(bytes))

        with self.assertRaises(TypeError) as error:
            NearIMU({"imu": "the_encoded_imu", "certainty": False})
        check_error_message(self, error, certainty_error_msg(bool))

        # valid calls

        NearIMU(
            {
                "imu": "test_imu",
            }
        )

        NearIMU({"imu": "test_imu_2", "certainty": 0.7})

    def test___str__(self):
        """
        Test the `__str__` method.
        """

        near_object = NearIMU(
            {
                "imu": "test_imu",
            }
        )
        self.assertEqual(str(near_object), 'nearIMU: {imu: "test_imu"} ')

        near_object = NearIMU({"imu": "test_imu", "certainty": 0.7})
        self.assertEqual(str(near_object), 'nearIMU: {imu: "test_imu" certainty: 0.7} ')



import sys
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, Union, cast
from typing_extensions import TypeAlias

import uuid as uuid_package

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_type_hints, get_origin, get_args
else:
    from typing import Annotated, get_type_hints, get_origin, get_args

from weaviate.collections.base import _CollectionBase
from weaviate.collections.classes.grpc import (
    FromNested,
    FromReference,
    FromReferenceMultiTarget,
    MetadataQuery,
    PROPERTIES,
    Generate,
)
from weaviate.collections.classes.types import Properties, P, TProperties
from weaviate.util import _to_beacons
from weaviate.types import UUIDS

from weaviate.proto.v1 import search_get_pb2


@dataclass
class _MetadataResult:
    uuid: Optional[uuid_package.UUID]
    vector: Optional[List[float]]
    creation_time_unix: Optional[int]
    last_update_time_unix: Optional[int]
    distance: Optional[float]
    certainty: Optional[float]
    score: Optional[float]
    explain_score: Optional[str]
    is_consistent: Optional[bool]
    generative: Optional[str]

    def _to_return(self) -> "_MetadataReturn":
        return _MetadataReturn(
            uuid=self.uuid,
            vector=self.vector,
            creation_time_unix=self.creation_time_unix,
            last_update_time_unix=self.last_update_time_unix,
            distance=self.distance,
            certainty=self.certainty,
            score=self.score,
            explain_score=self.explain_score,
            is_consistent=self.is_consistent,
        )


@dataclass
class _MetadataReturn:
    uuid: Optional[uuid_package.UUID]
    vector: Optional[List[float]]
    creation_time_unix: Optional[int]
    last_update_time_unix: Optional[int]
    distance: Optional[float]
    certainty: Optional[float]
    score: Optional[float]
    explain_score: Optional[str]
    is_consistent: Optional[bool]


@dataclass
class _Object(Generic[P]):
    properties: P
    metadata: _MetadataReturn


@dataclass
class _GroupByObject(Generic[P], _Object[P]):
    belongs_to_group: str


@dataclass
class _GenerativeObject(Generic[P], _Object[P]):
    generated: Optional[str]


@dataclass
class _GenerativeReturn(Generic[P]):
    objects: List[_GenerativeObject[P]]
    generated: Optional[str]


@dataclass
class _GroupByResult(Generic[P]):
    name: str
    min_distance: float
    max_distance: float
    number_of_objects: int
    objects: List[_Object[P]]


@dataclass
class _GroupByReturn(Generic[P]):
    objects: List[_GroupByObject[P]]
    groups: Dict[str, _GroupByResult]


@dataclass
class _QueryReturn(Generic[P]):
    objects: List[_Object[P]]


QueryReturn: TypeAlias = Union[_QueryReturn[Properties], _QueryReturn[TProperties]]
GenerativeReturn: TypeAlias = Union[_GenerativeReturn[Properties], _GenerativeReturn[TProperties]]
GroupByReturn: TypeAlias = Union[_GroupByReturn[Properties], _GroupByReturn[TProperties]]

ReturnProperties: TypeAlias = Union[PROPERTIES, Type[TProperties]]


class _Generative:
    single: Optional[str]
    grouped: Optional[str]
    grouped_properties: Optional[List[str]]

    def __init__(
        self,
        single: Optional[str],
        grouped: Optional[str],
        grouped_properties: Optional[List[str]],
    ) -> None:
        self.single = single
        self.grouped = grouped
        self.grouped_properties = grouped_properties

    def to_grpc(self) -> search_get_pb2.GenerativeSearch:
        return search_get_pb2.GenerativeSearch(
            single_response_prompt=self.single,
            grouped_response_task=self.grouped,
            grouped_properties=self.grouped_properties,
        )

    @classmethod
    def from_input(cls, generate: Optional[Generate]) -> Optional["_Generative"]:
        return (
            cls(
                single=generate.single_prompt,
                grouped=generate.grouped_task,
                grouped_properties=generate.grouped_properties,
            )
            if generate
            else None
        )


class _GroupBy:
    prop: str
    number_of_groups: int
    objects_per_group: int

    def __init__(self, prop: str, number_of_groups: int, objects_per_group: int) -> None:
        self.prop = prop
        self.number_of_groups = number_of_groups
        self.objects_per_group = objects_per_group

    def to_grpc(self) -> search_get_pb2.GroupBy:
        return search_get_pb2.GroupBy(
            path=[self.prop],
            number_of_groups=self.number_of_groups,
            objects_per_group=self.objects_per_group,
        )


Nested = Annotated[P, "NESTED"]


def __is_nested(value: Any) -> bool:
    return (
        get_origin(value) is Annotated
        and len(get_args(value)) == 2
        and cast(str, get_args(value)[1]) == "NESTED"
    )


def __create_nested_property_from_nested(name: str, value: Any) -> FromNested:
    if not __is_nested(value):
        raise ValueError(
            f"Non nested property detected in generic resolution, {value} of type {type(value)} is not allowed as a nested property."
        )
    inner_type = get_args(value)[0]
    return FromNested(
        name=name,
        properties=[
            __create_nested_property_from_nested(key, val) if __is_nested(val) else key
            for key, val in get_type_hints(inner_type, include_extras=True).items()
        ],
    )


class _Reference(Generic[P]):
    def __init__(
        self,
        objects: Optional[List[_Object[P]]],
        target_collection: Optional[str],
        uuids: Optional[UUIDS],
    ):
        """You should not initialise this class directly. Use the `.to()` or `.to_multi_target()` class methods instead."""
        self.__objects = objects
        self.__target_collection = target_collection if target_collection else ""
        self.__uuids = uuids

    def _to_beacons(self) -> List[Dict[str, str]]:
        if self.__uuids is None:
            return []
        return _to_beacons(self.__uuids, self.__target_collection)

    @classmethod
    def _from(cls, objects: List[_Object[P]]) -> "_Reference[P]":
        return cls(objects, None, None)

    @property
    def is_multi_target(self) -> bool:
        """Returns True if the reference is to a multi-target collection."""
        return self.__target_collection != ""

    @property
    def uuids_str(self) -> List[str]:
        """Returns the UUIDs as strings."""
        if isinstance(self.__uuids, list):
            return [str(uid) for uid in self.__uuids]
        else:
            return [str(self.__uuids)]

    @property
    def target_collection(self) -> str:
        """Returns the target collection name."""
        return self.__target_collection

    @property
    def objects(self) -> List[_Object[P]]:
        """Returns the objects of the cross reference."""
        return self.__objects or []


CrossReference = _Reference[P]
"""Use this TypeAlias when you want to type hint a cross reference within a generic data model.

If you want to define a reference property when creating your collection, use `ReferenceProperty` or `ReferencePropertyMultiTarget` instead.

If you want to create a reference when inserting an object, use `Reference.to()` or `Reference.to_multi_target()` instead.

Example:
    >>> from typing import TypedDict
    >>> import weaviate.classes as wvc
    >>>
    >>> class One(TypedDict):
    ...     prop: str
    >>>
    >>> class Two(TypedDict):
    ...     one: wvc.CrossReference[One]
"""


class Reference:
    """Factory class for cross references to other objects.

    Can be used with or without generics. If used with generics, the type of the cross reference can be defined from
    which the nested relationship will be used when performing queries using the generics. If used without generics,
    all returned objects will of the`Dict[str, Any]` type.
    """

    @classmethod
    def to(cls, uuids: UUIDS, data_model: Optional[Type[P]] = None) -> CrossReference[P]:
        """Define cross references to other objects by their UUIDs.

        Can be made to be generic by supplying a type to the `data_model` argument.

        Arguments:
            `uuids`
                List of UUIDs of the objects to which the reference should point.
        """
        return _Reference[P](None, None, uuids)

    @classmethod
    def to_multi_target(
        cls,
        uuids: UUIDS,
        target_collection: Union[str, _CollectionBase],
        data_model: Optional[Type[P]] = None,
    ) -> CrossReference[P]:
        """Define cross references to other objects by their UUIDs and the collection in which they are stored.

        Can be made to be generic by supplying a type to the `data_model` argument.

        Arguments:
            `uuids`
                List of UUIDs of the objects to which the reference should point.
            `target_collection`
                The collection in which the objects are stored. Can be either the name of the collection or the collection object itself.
        """
        return _Reference[P](
            None,
            target_collection.name
            if isinstance(target_collection, _CollectionBase)
            else target_collection,
            uuids,
        )


def _metadata_from_dict(metadata: Dict[str, Any]) -> _MetadataReturn:
    return _MetadataReturn(
        uuid=uuid_package.UUID(metadata["id"]) if "id" in metadata else None,
        vector=metadata.get("vector"),
        creation_time_unix=metadata.get("creationTimeUnix"),
        last_update_time_unix=metadata.get("lastUpdateTimeUnix"),
        distance=metadata.get("distance"),
        certainty=metadata.get("certainty"),
        explain_score=metadata.get("explainScore"),
        score=metadata.get("score"),
        is_consistent=metadata.get("isConsistent"),
    )


def _extract_property_type_from_reference(type_: _Reference[P]) -> Type[P]:
    """Extract inner type from Reference[Properties]."""
    if getattr(type_, "__origin__", None) == _Reference:
        args = cast(List[Type[P]], getattr(type_, "__args__", None))
        return args[0]
    raise ValueError("Type is not Reference[Properties]")


def _extract_property_type_from_annotated_reference(
    type_: Union[
        Annotated[_Reference[P], MetadataQuery],
        Annotated[_Reference[P], MetadataQuery, str],
    ]
) -> Type[P]:
    """Extract inner type from Annotated[Reference[Properties]]."""
    if get_origin(type_) is Annotated:
        args = cast(List[_Reference[Type[P]]], getattr(type_, "__args__", None))
        inner_type = args[0]
        if get_origin(inner_type) is _Reference:
            inner_args = cast(List[Type[P]], getattr(inner_type, "__args__", None))
            return inner_args[0]
    raise ValueError("Type is not Annotated[Reference[Properties]]")


def __is_annotated_reference(value: Any) -> bool:
    return (
        get_origin(value) is Annotated
        and len(get_args(value)) == 2
        and get_origin(get_args(value)[0]) is _Reference
    )


def __create_link_to_from_annotated_reference(
    link_on: str,
    value: Union[
        Annotated[_Reference[Properties], MetadataQuery],
        Annotated[_Reference[Properties], MetadataQuery, str],
    ],
) -> Union[FromReference, FromReferenceMultiTarget]:
    """Create FromReference or FromReferenceMultiTarget from Annotated[Reference[Properties]]."""
    assert get_origin(value) is Annotated
    args = cast(List[_Reference[Properties]], getattr(value, "__args__", None))
    inner_type = args[0]
    assert get_origin(inner_type) is _Reference
    inner_type_metadata = cast(
        Union[Tuple[MetadataQuery], Tuple[MetadataQuery, str]], getattr(value, "__metadata__", None)
    )
    metadata = inner_type_metadata[0]
    if len(inner_type_metadata) == 2:
        target_collection = cast(Tuple[MetadataQuery, str], inner_type_metadata)[
            1
        ]  # https://github.com/python/mypy/issues/1178
        return FromReferenceMultiTarget(
            link_on=link_on,
            return_metadata=metadata,
            return_properties=_extract_properties_from_data_model(
                _extract_property_type_from_annotated_reference(value)
            ),
            target_collection=target_collection,
        )
    else:
        return FromReference(
            link_on=link_on,
            return_metadata=metadata,
            return_properties=_extract_properties_from_data_model(
                _extract_property_type_from_annotated_reference(value)
            ),
        )


def __is_reference(value: Any) -> bool:
    return get_origin(value) is _Reference


def __create_link_to_from_reference(
    link_on: str,
    value: _Reference[Properties],
) -> FromReference:
    """Create FromReference from Reference[Properties]."""
    return FromReference(
        link_on=link_on,
        return_metadata=MetadataQuery(
            uuid=True,
            creation_time_unix=True,
            last_update_time_unix=True,
            distance=True,
            certainty=True,
            score=True,
            explain_score=True,
            is_consistent=True,
        ),
        return_properties=_extract_properties_from_data_model(
            _extract_property_type_from_reference(value)
        ),
    )


def _extract_properties_from_data_model(type_: Type[Properties]) -> PROPERTIES:
    """Extract properties of Properties recursively from Properties.

    Checks to see if there is a _Reference[Properties], Annotated[_Reference[Properties]], or _Nested[Properties]
    in the data model and lists out the properties as classes readily consumable by the underlying API.
    """
    return [
        __create_link_to_from_annotated_reference(key, value)
        if __is_annotated_reference(value)
        else (
            __create_link_to_from_reference(key, value)
            if __is_reference(value)
            else (__create_nested_property_from_nested(key, value) if __is_nested(value) else key)
        )
        for key, value in get_type_hints(type_, include_extras=True).items()
    ]



from enum import Enum
from pydantic import BaseModel, ConfigDict, Field


class TenantActivityStatus(str, Enum):
    """TenantActivityStatus class used to describe the activity status of a tenant in Weaviate.

    Attributes:
        `HOT`
            The tenant is fully active and can be used.
        `COLD`
            The tenant is not active, files stored locally.
    """

    HOT = "HOT"
    COLD = "COLD"


class Tenant(BaseModel):
    """Tenant class used to describe a tenant in Weaviate.

    Attributes:
        `name`
            the name of the tenant.
        `activity_status`
            TenantActivityStatus, default: "HOT"
    """

    model_config = ConfigDict(populate_by_name=True)
    name: str
    activityStatus: TenantActivityStatus = Field(
        default=TenantActivityStatus.HOT, alias="activity_status"
    )

    @property
    def activity_status(self) -> TenantActivityStatus:
        """Getter for the activity status of the tenant."""
        return self.activityStatus



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from typing_extensions import TypeAlias

import grpc  # type: ignore
import uuid as uuid_lib
from google.protobuf import struct_pb2

from weaviate.collections.classes.config import ConsistencyLevel

from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import (
    HybridFusion,
    FromReference,
    FromReferenceMultiTarget,
    FromNested,
    MetadataQuery,
    Move,
    PROPERTIES,
    PROPERTY,
    Sort,
)
from weaviate.collections.classes.internal import _Generative, _GroupBy
from weaviate.collections.filters import _FilterToGRPC

from weaviate.collections.grpc.shared import _BaseGRPC

from weaviate.connect import Connection
from weaviate.exceptions import WeaviateQueryException
from weaviate.types import UUID

from weaviate.proto.v1 import search_get_pb2


# Can be found in the google.protobuf.internal.well_known_types.pyi stub file but is defined explicitly here for clarity.
_StructValue: TypeAlias = Union[
    struct_pb2.Struct,
    struct_pb2.ListValue,
    str,
    float,
    bool,
    None,
    List[float],
    List[int],
    List[str],
    List[bool],
    List[UUID],
]
_PyValue: TypeAlias = Union[
    Dict[str, "_PyValue"],
    List["_PyValue"],
    str,
    float,
    bool,
    None,
    List[float],
    List[int],
    List[str],
    List[bool],
    List[UUID],
]


@dataclass
class SearchResult:
    """Represents a single search result from Weaviate."""

    properties: search_get_pb2.PropertiesResult
    metadata: search_get_pb2.MetadataResult


@dataclass
class GroupByResult:
    """Represents a single group-by result from Weaviate."""

    name: str
    min_distance: float
    max_distance: float
    number_of_objects: int
    objects: List[SearchResult]


@dataclass
class SearchResponse:
    """Represents the response from a search request to Weaviate."""

    # the name of these members must match the proto file
    results: List[SearchResult]
    generative_grouped_result: str
    group_by_results: List[GroupByResult]


@dataclass
class _Move:
    force: float
    concepts: List[str]
    objects: List[uuid_lib.UUID]


class _QueryGRPC(_BaseGRPC):
    def __init__(
        self,
        connection: Connection,
        name: str,
        tenant: Optional[str],
        consistency_level: Optional[ConsistencyLevel],
        default_properties: Optional[PROPERTIES] = None,
    ):
        super().__init__(connection, consistency_level)
        self._name: str = name
        self._tenant = tenant

        if default_properties is not None:
            self._default_props: Set[PROPERTY] = self.__convert_properties_to_set(
                default_properties
            )
        else:
            self._default_props = set()
        self._metadata: Optional[MetadataQuery] = None

        self._limit: Optional[int] = None
        self._offset: Optional[int] = None
        self._autocut: Optional[int] = None
        self._after: Optional[UUID] = None

        self._hybrid_query: Optional[str] = None
        self._hybrid_alpha: Optional[float] = None
        self._hybrid_vector: Optional[List[float]] = None
        self._hybrid_properties: Optional[List[str]] = None
        self._hybrid_fusion_type: Optional[int] = None

        self._bm25_query: Optional[str] = None
        self._bm25_properties: Optional[List[str]] = None

        self._near_vector_vec: Optional[List[float]] = None
        self._near_object_obj: Optional[UUID] = None
        self._near_text: Optional[List[str]] = None
        self._near_text_move_away: Optional[search_get_pb2.NearTextSearch.Move] = None
        self._near_text_move_to: Optional[search_get_pb2.NearTextSearch.Move] = None

        self._near_certainty: Optional[float] = None
        self._near_distance: Optional[float] = None

        self._near_image: Optional[str] = None
        self._near_video: Optional[str] = None
        self._near_audio: Optional[str] = None

        self._generative: Optional[_Generative] = None

        self._sort: Optional[List[Sort]] = None

        self._group_by: Optional[_GroupBy] = None

        self._filters: Optional[_Filters] = None

    def __parse_sort(self, sort: Optional[Union[Sort, List[Sort]]]) -> None:
        if sort is None:
            self._sort = None
        elif isinstance(sort, Sort):
            self._sort = [sort]
        else:
            self._sort = sort

    def get(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Union[Sort, List[Sort]]] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
        generative: Optional[_Generative] = None,
    ) -> SearchResponse:
        self._limit = limit
        self._offset = offset
        self._after = after
        self._filters = filters
        self._metadata = return_metadata
        self.__parse_sort(sort)
        self.__merge_default_and_return_properties(return_properties)
        self._generative = generative
        return self.__call()

    def hybrid(
        self,
        query: str,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
        generative: Optional[_Generative] = None,
    ) -> SearchResponse:
        self._hybrid_query = query
        self._hybrid_alpha = alpha
        self._hybrid_vector = vector
        self._hybrid_properties = properties
        self._hybrid_fusion_type = (
            search_get_pb2.Hybrid.FusionType.Value(fusion_type.value)
            if fusion_type is not None
            else None
        )
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)

        self._generative = generative

        return self.__call()

    def bm25(
        self,
        query: str,
        properties: Optional[List[str]] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
        generative: Optional[_Generative] = None,
    ) -> SearchResponse:
        self._bm25_query = query
        self._bm25_properties = properties
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)

        self._generative = generative

        return self.__call()

    def near_vector(
        self,
        near_vector: List[float],
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> SearchResponse:
        self._near_vector_vec = near_vector
        self._near_certainty = certainty
        self._near_distance = distance
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._metadata = return_metadata
        self._group_by = group_by
        self._generative = generative
        self.__merge_default_and_return_properties(return_properties)

        return self.__call()

    def near_object(
        self,
        near_object: UUID,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> SearchResponse:
        self._near_object_obj = near_object
        self._near_certainty = certainty
        self._near_distance = distance
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)
        self._group_by = group_by
        self._generative = generative
        return self.__call()

    def near_text(
        self,
        near_text: Union[List[str], str],
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        move_to: Optional[Move] = None,
        move_away: Optional[Move] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> SearchResponse:
        if isinstance(near_text, str):
            near_text = [near_text]
        self._near_text = near_text
        self._near_certainty = certainty
        self._near_distance = distance
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        if move_away is not None:
            self._near_text_move_away = search_get_pb2.NearTextSearch.Move(
                force=move_away.force,
                concepts=move_away._concepts_list,
                uuids=move_away._objects_list,
            )
        if move_to is not None:
            self._near_text_move_to = search_get_pb2.NearTextSearch.Move(
                force=move_to.force, concepts=move_to._concepts_list, uuids=move_to._objects_list
            )

        self._generative = generative

        self._group_by = group_by
        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)

        return self.__call()

    def near_image(
        self,
        image: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> SearchResponse:
        self._near_image = image
        self._near_certainty = certainty
        self._near_distance = distance
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._group_by = group_by
        self._generative = generative

        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)

        return self.__call()

    def near_video(
        self,
        video: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> SearchResponse:
        self._near_video = video
        self._near_certainty = certainty
        self._near_distance = distance
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._group_by = group_by
        self._generative = generative

        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)

        return self.__call()

    def near_audio(
        self,
        audio: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        autocut: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[_GroupBy] = None,
        generative: Optional[_Generative] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> SearchResponse:
        self._near_audio = audio
        self._near_certainty = certainty
        self._near_distance = distance
        self._limit = limit
        self._autocut = autocut
        self._filters = filters
        self._group_by = group_by
        self._generative = generative

        self._metadata = return_metadata
        self.__merge_default_and_return_properties(return_properties)

        return self.__call()

    def __call(self) -> SearchResponse:
        metadata: Optional[Tuple[Tuple[str, str], ...]] = None
        access_token = self._connection.get_current_bearer_token()

        metadata_list: List[Tuple[str, str]] = []
        if len(access_token) > 0:
            metadata_list.append(("authorization", access_token))

        if len(self._connection.additional_headers):
            for key, val in self._connection.additional_headers.items():
                if val is not None:
                    metadata_list.append((key.lower(), val))

        if len(metadata_list) > 0:
            metadata = tuple(metadata_list)

        try:
            print(
                self._translate_properties_from_python_to_grpc(self._default_props),
            )
            assert self._connection.grpc_stub is not None
            res: SearchResponse  # According to PEP-0526
            res, _ = self._connection.grpc_stub.Search.with_call(
                search_get_pb2.SearchRequest(
                    collection=self._name,
                    limit=self._limit,
                    offset=self._offset,
                    after=str(self._after) if self._after is not None else "",
                    autocut=self._autocut,
                    near_vector=search_get_pb2.NearVector(
                        vector=self._near_vector_vec,
                        certainty=self._near_certainty,
                        distance=self._near_distance,
                    )
                    if self._near_vector_vec is not None
                    else None,
                    near_object=search_get_pb2.NearObject(
                        id=str(self._near_object_obj),
                        certainty=self._near_certainty,
                        distance=self._near_distance,
                    )
                    if self._near_object_obj is not None
                    else None,
                    properties=self._translate_properties_from_python_to_grpc(self._default_props),
                    metadata=self._metadata_to_grpc(self._metadata)
                    if self._metadata is not None
                    else None,
                    bm25_search=search_get_pb2.BM25(
                        properties=self._bm25_properties, query=self._bm25_query
                    )
                    if self._bm25_query is not None
                    else None,
                    hybrid_search=search_get_pb2.Hybrid(
                        properties=self._hybrid_properties,
                        query=self._hybrid_query,
                        alpha=self._hybrid_alpha,
                        vector=self._hybrid_vector,
                        fusion_type=cast(
                            search_get_pb2.Hybrid.FusionType, self._hybrid_fusion_type
                        ),
                    )
                    if self._hybrid_query is not None
                    else None,
                    tenant=self._tenant,
                    filters=_FilterToGRPC.convert(self._filters),
                    near_text=search_get_pb2.NearTextSearch(
                        query=self._near_text,
                        certainty=self._near_certainty,
                        distance=self._near_distance,
                        move_to=self._near_text_move_to,
                        move_away=self._near_text_move_away,
                    )
                    if self._near_text is not None
                    else None,
                    near_image=search_get_pb2.NearImageSearch(
                        image=self._near_image,
                        distance=self._near_distance,
                        certainty=self._near_certainty,
                    )
                    if self._near_image is not None
                    else None,
                    near_video=search_get_pb2.NearVideoSearch(
                        video=self._near_video,
                        distance=self._near_distance,
                        certainty=self._near_certainty,
                    )
                    if self._near_video is not None
                    else None,
                    near_audio=search_get_pb2.NearAudioSearch(
                        audio=self._near_audio,
                        distance=self._near_distance,
                        certainty=self._near_certainty,
                    )
                    if self._near_audio is not None
                    else None,
                    consistency_level=self._consistency_level,
                    sort_by=[
                        search_get_pb2.SortBy(ascending=sort.ascending, path=[sort.prop])
                        for sort in self._sort
                    ]
                    if self._sort is not None
                    else None,
                    generative=self._generative.to_grpc() if self._generative is not None else None,
                    group_by=self._group_by.to_grpc() if self._group_by is not None else None,
                ),
                metadata=metadata,
            )

            return res

        except grpc.RpcError as e:
            raise WeaviateQueryException(e.details())

    def _metadata_to_grpc(self, metadata: MetadataQuery) -> search_get_pb2.MetadataRequest:
        return search_get_pb2.MetadataRequest(
            uuid=metadata.uuid,
            vector=metadata.vector,
            creation_time_unix=metadata.creation_time_unix,
            last_update_time_unix=metadata.last_update_time_unix,
            distance=metadata.distance,
            certainty=metadata.certainty,
            explain_score=metadata.explain_score,
            score=metadata.score,
            is_consistent=metadata.is_consistent,
        )

    def _translate_properties_from_python_to_grpc(
        self, properties: Set[PROPERTY]
    ) -> search_get_pb2.PropertiesRequest:
        def resolve_property(prop: FromNested) -> search_get_pb2.ObjectPropertiesRequest:
            props = prop.properties if isinstance(prop.properties, list) else [prop.properties]
            return search_get_pb2.ObjectPropertiesRequest(
                prop_name=prop.name,
                primitive_properties=[p for p in props if isinstance(p, str)],
                object_properties=[resolve_property(p) for p in props if isinstance(p, FromNested)],
            )

        return search_get_pb2.PropertiesRequest(
            non_ref_properties=[prop for prop in properties if isinstance(prop, str)],
            ref_properties=[
                search_get_pb2.RefPropertiesRequest(
                    reference_property=prop.link_on,
                    properties=self._translate_properties_from_python_to_grpc(
                        self.__convert_properties_to_set(prop.return_properties)
                    )
                    if prop.return_properties is not None
                    else None,
                    metadata=self._metadata_to_grpc(prop.return_metadata)
                    if prop.return_metadata is not None
                    else None,
                    target_collection=prop.target_collection
                    if isinstance(prop, FromReferenceMultiTarget)
                    else None,
                )
                for prop in properties
                if isinstance(prop, FromReference)
            ],
            object_properties=[
                resolve_property(prop) for prop in properties if isinstance(prop, FromNested)
            ],
        )

    def __merge_default_and_return_properties(
        self, return_properties: Optional[PROPERTIES]
    ) -> None:
        if return_properties is not None:
            self._default_props = self._default_props.union(
                self.__convert_properties_to_set(return_properties)
            )

    @staticmethod
    def __convert_properties_to_set(properties: PROPERTIES) -> Set[PROPERTY]:
        if isinstance(properties, list):
            return set(properties)
        else:
            return {properties}



"""
Weaviate Exceptions.
"""

from requests import Response, exceptions

ERROR_CODE_EXPLANATION = {
    413: """Payload Too Large. Try to decrease the batch size or increase the maximum request size on your weaviate
         server."""
}


class WeaviateBaseError(Exception):
    """
    Weaviate base exception that all Weaviate exceptions should inherit from.
    This error can be used to catch any Weaviate exceptions.
    """

    def __init__(self, message: str = ""):
        """
        Weaviate base exception initializer.
        Parameters
        ----------
        message: str, optional
            An error message specific to the context in which the error occurred.
        """

        self.message = message
        super().__init__(message)


class UnexpectedStatusCodeException(WeaviateBaseError):
    """
    Is raised in case the status code returned from Weaviate is
    not handled in the client implementation and suggests an error.
    """

    def __init__(self, message: str, response: Response):
        """
        Is raised in case the status code returned from Weaviate is
        not handled in the client implementation and suggests an error.

        Custom code can act on the attributes:
        - status_code
        - json

        Parameters
        ----------
        message: str
            An error message specific to the context, in which the error occurred.
        response: requests.Response
            The request response of which the status code was unexpected.
        """
        self._status_code: int = response.status_code
        # Set error message

        try:
            body = response.json()
        except exceptions.JSONDecodeError:
            body = None

        msg = (
            message
            + f"! Unexpected status code: {response.status_code}, with response body: {body}."
        )
        if response.status_code in ERROR_CODE_EXPLANATION:
            msg += " " + ERROR_CODE_EXPLANATION[response.status_code]

        super().__init__(msg)

    @property
    def status_code(self) -> int:
        return self._status_code


class ResponseCannotBeDecodedException(WeaviateBaseError):
    def __init__(self, location: str, response: Response):
        """Raised when a weaviate response cannot be decoded to json

        Parameters
        ----------
        location: str
            From which code path the exception was raised.
        response: requests.Response
            The request response of which the status code was unexpected.
        """
        msg = f"Cannot decode response from weaviate {response} with content {response.text} for request from {location}"
        super().__init__(msg)
        self._status_code: int = response.status_code

    @property
    def status_code(self) -> int:
        return self._status_code


class ObjectAlreadyExistsException(WeaviateBaseError):
    """
    Object Already Exists Exception.
    """


class AuthenticationFailedException(WeaviateBaseError):
    """
    Authentication Failed Exception.
    """


class SchemaValidationException(WeaviateBaseError):
    """
    Schema Validation Exception.
    """


class BackupFailedException(WeaviateBaseError):
    """
    Backup Failed Exception.
    """


class EmptyResponseException(WeaviateBaseError):
    """
    Occurs when an HTTP request unexpectedly returns an empty response
    """


class MissingScopeException(WeaviateBaseError):
    """Scope was not provided with client credential flow."""


class AdditionalPropertiesException(WeaviateBaseError):
    """Additional properties were provided multiple times."""

    def __init__(self, additional_dict: str, additional_dataclass: str):
        msg = f"""
        Cannot add AdditionalProperties class together with string-additional properties. Did you call
            .with_additional() multiple times?.
             Current additional properties already present:
                - strings: {additional_dict}
                - AdditionalProperties class: {additional_dataclass}
        """
        super().__init__(msg)


class InvalidDataModelException(WeaviateBaseError):
    """Is raised when the user provides a generic that is not supported"""

    def __init__(self) -> None:
        msg = """data_model can only be a dict type, e.g. Dict[str, str], or a class that inherits from TypedDict"""
        super().__init__(msg)


class WeaviateStartUpError(WeaviateBaseError):
    """Is raised if weaviate does not start up in time."""


class WeaviateEmbeddedInvalidVersion(WeaviateBaseError):
    """Invalid version provided to Weaviate embedded."""

    def __init__(self, url: str):
        msg = f"""Invalid version provided to Weaviate embedded. It must be either:
        - a url to a tar.gz file that contains a Weaviate binary
        - a version number, eg "1.18.2"
        - the string "latest" to download the latest non-beta version

        Url provided was: {url}.
        """
        super().__init__(msg)


class WeaviateInvalidInputException(WeaviateBaseError):
    """Is raised if the input to a function is invalid."""

    def __init__(self, message: str):
        msg = f"""Invalid input provided: {message}."""
        super().__init__(msg)
        self.message = message


class WeaviateQueryException(WeaviateBaseError):
    """Is raised if a query (either gRPC or GraphQL) to Weaviate fails in any way."""

    def __init__(self, message: str):
        msg = f"""Query call failed with message {message}."""
        super().__init__(msg)
        self.message = message


class WeaviateAddInvalidPropertyError(WeaviateBaseError):
    """Is raised when adding an invalid new property."""

    def __init__(self, message: str):
        msg = f"""Could not add the property {message}. Only optional properties or properties with default
        value are valid"""
        super().__init__(msg)
        self.message = message


class WeaviateBatchValidationError(WeaviateBaseError):
    """Is raised when a batch validation error occurs."""

    def __init__(self, message: str):
        msg = f"""Batch validation error: {message}"""
        super().__init__(msg)
        self.message = message


class WeaviateInsertInvalidPropertyError(WeaviateBaseError):
    """Is raised when inserting an invalid property."""

    def __init__(self, data: dict):
        msg = f"""It is forbidden to insert `id` or `vector` inside properties: {data}. Only properties defined in your collection's config can be insterted as properties of the object, `id` is totally forbidden as it is reserved and `vector` is forbidden at this level. You should use the `DataObject` class if you wish to insert an object with a custom `vector` whilst inserting its properties."""
        super().__init__(msg)


class WeaviateGrpcUnavailable(WeaviateBaseError):
    """Is raised when a gRPC-backed query is made with no gRPC connection present."""

    def __init__(self) -> None:
        msg = """gRPC is not available. Please make sure that gRPC is configured correctly in the client and on the server."""
        super().__init__(msg)



import datetime
import uuid
from typing import List

import pytest as pytest

import weaviate
from weaviate.collections.classes.config import (
    Configure,
    Property,
    DataType,
    ReferenceProperty,
    ReferencePropertyMultiTarget,
    Tokenization,
)
from weaviate.collections.classes.data import DataObject
from weaviate.collections.classes.filters import (
    Filter,
    _Filters,
    _FilterValue,
)
from weaviate.collections.classes.grpc import MetadataQuery
from weaviate.collections.classes.internal import Reference

NOW = datetime.datetime.now(datetime.timezone.utc)
LATER = NOW + datetime.timedelta(hours=1)
MUCH_LATER = NOW + datetime.timedelta(days=1)

UUID1 = uuid.uuid4()
UUID2 = uuid.uuid4()
UUID3 = uuid.uuid4()


@pytest.fixture(scope="module")
def client():
    client = weaviate.connect_to_local()
    client.collections.delete_all()
    yield client
    client.collections.delete_all()


@pytest.mark.parametrize(
    "weaviate_filter,results",
    [
        (Filter(path="name").equal("Banana"), [0]),
        (Filter(path="name").not_equal("Banana"), [1, 2]),
        (Filter(path="name").like("*nana"), [0]),
    ],
)
def test_filters_text(
    client: weaviate.WeaviateClient, weaviate_filter: _FilterValue, results: List[int]
):
    client.collections.delete("TestFilterText")
    collection = client.collections.create(
        name="TestFilterText",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="name", data_type=DataType.TEXT)],
    )

    uuids = [
        collection.data.insert({"name": "Banana"}),
        collection.data.insert({"name": "Apple"}),
        collection.data.insert({"name": "Mountain"}),
    ]

    objects = collection.query.fetch_objects(filters=weaviate_filter).objects
    assert len(objects) == len(results)

    uuids = [uuids[result] for result in results]
    assert all(obj.metadata.uuid in uuids for obj in objects)


@pytest.mark.parametrize(
    "weaviate_filter,results",
    [
        (Filter(path="num").greater_than(1) & Filter(path="num").less_than(3), [1]),
        (
            (Filter(path="num").less_or_equal(1)) | Filter(path="num").greater_or_equal(3),
            [0, 2],
        ),
        (
            Filter(path="num").less_or_equal(1) | Filter(path="num").greater_or_equal(3),
            [0, 2],
        ),
        (
            (Filter(path="num").less_or_equal(1) & Filter(path="num").greater_or_equal(1))
            | Filter(path="num").greater_or_equal(3)
            | Filter(path="num").is_none(True),
            [0, 2, 3],
        ),
    ],
)
def test_filters_nested(
    client: weaviate.WeaviateClient,
    weaviate_filter: _Filters,
    results: List[int],
):
    client.collections.delete("TestFilterNested")
    collection = client.collections.create(
        name="TestFilterNested",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="num", data_type=DataType.NUMBER)],
        inverted_index_config=Configure.inverted_index(index_null_state=True),
    )

    uuids = [
        collection.data.insert({"num": 1.0}),
        collection.data.insert({"num": 2.0}),
        collection.data.insert({"num": 3.0}),
        collection.data.insert({"num": None}),
    ]

    objects = collection.query.fetch_objects(
        filters=weaviate_filter, return_metadata=MetadataQuery(uuid=True)
    ).objects
    assert len(objects) == len(results)

    uuids = [uuids[result] for result in results]
    assert all(obj.metadata.uuid in uuids for obj in objects)


def test_length_filter(client: weaviate.WeaviateClient):
    client.collections.delete("TestFilterNested")
    collection = client.collections.create(
        name="TestFilterNested",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="field", data_type=DataType.TEXT)],
        inverted_index_config=Configure.inverted_index(index_property_length=True),
    )
    uuids = [
        collection.data.insert({"field": "one"}),
        collection.data.insert({"field": "two"}),
        collection.data.insert({"field": "three"}),
        collection.data.insert({"field": "four"}),
    ]
    objects = collection.query.fetch_objects(
        filters=Filter(path="field", length=True).equal(3)
    ).objects

    results = [0, 1]
    assert len(objects) == len(results)
    uuids = [uuids[result] for result in results]
    assert all(obj.metadata.uuid in uuids for obj in objects)


@pytest.mark.parametrize(
    "weaviate_filter,results",
    [
        (Filter(path="number").is_none(True), [3]),
        (Filter(path="number").is_none(False), [0, 1, 2]),
    ],
)
def test_filters_comparison(
    client: weaviate.WeaviateClient, weaviate_filter: _FilterValue, results: List[int]
):
    client.collections.delete("TestFilterNumber")
    collection = client.collections.create(
        name="TestFilterNumber",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="number", data_type=DataType.INT)],
        inverted_index_config=Configure.inverted_index(index_null_state=True),
    )

    uuids = [
        collection.data.insert({"number": 1}),
        collection.data.insert({"number": 2}),
        collection.data.insert({"number": 3}),
        collection.data.insert({"number": None}),
    ]

    objects = collection.query.fetch_objects(filters=weaviate_filter).objects
    assert len(objects) == len(results)

    uuids = [uuids[result] for result in results]
    assert all(obj.metadata.uuid in uuids for obj in objects)


@pytest.mark.parametrize(
    "weaviate_filter,results",
    [
        (Filter(path="nums").contains_any([1, 4]), [0, 3]),
        (Filter(path="nums").contains_any([10]), []),
        (Filter(path="num").contains_any([1]), [0, 1]),
        (Filter(path="text").contains_any(["test"]), [0, 1]),
        (Filter(path="text").contains_any(["real", "deal"]), [1, 2, 3]),
        (Filter(path="texts").contains_any(["test"]), [0, 1]),
        (Filter(path="texts").contains_any(["real", "deal"]), [1, 2, 3]),
        (Filter(path="float").contains_any([2.0]), []),
        (Filter(path="float").contains_any([2]), []),
        (Filter(path="float").contains_any([8]), [3]),
        (Filter(path="float").contains_any([8.0]), [3]),
        (Filter(path="floats").contains_any([2.0]), [0, 1]),
        (Filter(path="floats").contains_any([0.4, 0.7]), [0, 1, 3]),
        (Filter(path="floats").contains_any([2]), [0, 1]),
        (Filter(path="bools").contains_any([True, False]), [0, 1, 3]),
        (Filter(path="bools").contains_any([False]), [0, 1]),
        (Filter(path="bool").contains_any([True]), [0, 1, 3]),
        (Filter(path="nums").contains_all([1, 4]), [0]),
        (Filter(path="text").contains_all(["real", "test"]), [1]),
        (Filter(path="texts").contains_all(["real", "test"]), [1]),
        (Filter(path="floats").contains_all([0.7, 2]), [1]),
        (Filter(path="bools").contains_all([True, False]), [0]),
        (Filter(path="bool").contains_all([True, False]), []),
        (Filter(path="bool").contains_all([True]), [0, 1, 3]),
        (Filter(path="dates").contains_any([NOW, MUCH_LATER]), [0, 1, 3]),
        (Filter(path="dates").contains_any([NOW]), [0, 1]),
        (Filter(path="date").equal(NOW), [0]),
        (Filter(path="date").greater_than(NOW), [1, 3]),
        (Filter(path="uuids").contains_all([UUID2, UUID1]), [0, 3]),
        (Filter(path="uuids").contains_any([UUID2, UUID1]), [0, 1, 3]),
        (Filter(path="uuid").contains_any([UUID3]), []),
        (Filter(path="uuid").contains_any([UUID1]), [0]),
    ],
)
def test_filters_contains(
    client: weaviate.WeaviateClient, weaviate_filter: _FilterValue, results: List[int]
):
    client.collections.delete("TestFilterContains")
    collection = client.collections.create(
        name="TestFilterContains",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(name="texts", data_type=DataType.TEXT_ARRAY),
            Property(name="num", data_type=DataType.INT),
            Property(name="nums", data_type=DataType.INT_ARRAY),
            Property(name="float", data_type=DataType.NUMBER),
            Property(name="floats", data_type=DataType.NUMBER_ARRAY),
            Property(name="bool", data_type=DataType.BOOL),
            Property(name="bools", data_type=DataType.BOOL_ARRAY),
            Property(name="dates", data_type=DataType.DATE_ARRAY),
            Property(name="date", data_type=DataType.DATE),
            Property(name="uuids", data_type=DataType.UUID_ARRAY),
            Property(name="uuid", data_type=DataType.UUID),
        ],
    )

    uuids = [
        collection.data.insert(
            {
                "text": "this is a test",
                "texts": "this is a test".split(" "),
                "num": 1,
                "nums": [1, 2, 4],
                "float": 0.5,
                "floats": [0.4, 0.9, 2],
                "bool": True,
                "bools": [True, False],
                "dates": [NOW, LATER, MUCH_LATER],
                "date": NOW,
                "uuids": [UUID1, UUID3, UUID2],
                "uuid": UUID1,
            }
        ),
        collection.data.insert(
            {
                "text": "this is not a real test",
                "texts": "this is not a real test".split(" "),
                "num": 1,
                "nums": [5, 6, 9],
                "float": 0.3,
                "floats": [0.1, 0.7, 2],
                "bool": True,
                "bools": [False, False],
                "dates": [NOW, NOW, MUCH_LATER],
                "date": LATER,
                "uuids": [UUID2, UUID2],
                "uuid": UUID2,
            }
        ),
        collection.data.insert(
            {
                "text": "real deal",
                "texts": "real deal".split(" "),
                "num": 3,
                "nums": [],
                "floats": [],
                "bool": False,
                "bools": [],
                "dates": [],
                "uuids": [],
            }
        ),
        collection.data.insert(
            {
                "text": "not real deal",
                "texts": "not real deal".split(" "),
                "num": 4,
                "nums": [4],
                "float": 8,
                "floats": [0.7],
                "bool": True,
                "bools": [True],
                "dates": [MUCH_LATER],
                "date": MUCH_LATER,
                "uuids": [UUID1, UUID2],
                "uuid": UUID2,
            }
        ),
    ]

    objects = collection.query.fetch_objects(
        filters=weaviate_filter, return_metadata=MetadataQuery(uuid=True)
    ).objects
    assert len(objects) == len(results)

    uuids = [uuids[result] for result in results]
    assert all(obj.metadata.uuid in uuids for obj in objects)


@pytest.mark.parametrize(
    "weaviate_filter,results",
    [
        (Filter(path=["ref", "TestFilterRef2", "int"]).greater_than(3), [1]),
        (Filter(path=["ref", "TestFilterRef2", "text"], length=True).less_than(6), [0]),
    ],
)
def test_ref_filters(
    client: weaviate.WeaviateClient, weaviate_filter: _FilterValue, results: List[int]
):
    client.collections.delete("TestFilterRef")
    client.collections.delete("TestFilterRef2")
    to_collection = client.collections.create(
        name="TestFilterRef2",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="int", data_type=DataType.INT),
            Property(name="text", data_type=DataType.TEXT),
        ],
        inverted_index_config=Configure.inverted_index(index_property_length=True),
    )
    uuids_to = [
        to_collection.data.insert(properties={"int": 0, "text": "first"}),
        to_collection.data.insert(properties={"int": 15, "text": "second"}),
    ]
    from_collection = client.collections.create(
        name="TestFilterRef",
        properties=[
            ReferenceProperty(name="ref", target_collection="TestFilterRef2"),
            Property(name="name", data_type=DataType.TEXT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    uuids_from = [
        from_collection.data.insert({"ref": Reference.to(uuids_to[0]), "name": "first"}),
        from_collection.data.insert({"ref": Reference.to(uuids_to[1]), "name": "second"}),
    ]

    objects = from_collection.query.fetch_objects(
        filters=weaviate_filter, return_metadata=MetadataQuery(uuid=True)
    ).objects
    assert len(objects) == len(results)

    uuids = [uuids_from[result] for result in results]
    assert all(obj.metadata.uuid in uuids for obj in objects)


def test_ref_filters_multi_target(client: weaviate.WeaviateClient):
    target = "TestFilterRefMulti2"
    source = "TestFilterRefMulti"
    client.collections.delete(source)
    client.collections.delete(target)
    to_collection = client.collections.create(
        name=target,
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="int", data_type=DataType.INT)],
    )
    uuid_to = to_collection.data.insert(properties={"int": 0})
    uuid_to2 = to_collection.data.insert(properties={"int": 5})
    from_collection = client.collections.create(
        name=source,
        properties=[
            ReferencePropertyMultiTarget(
                name="ref", target_collections=[target, "TestFilterRefMulti"]
            ),
            Property(name="name", data_type=DataType.TEXT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    uuid_from_to_target1 = from_collection.data.insert(
        {
            "ref": Reference.to_multi_target(uuids=uuid_to, target_collection=target),
            "name": "first",
        }
    )
    uuid_from_to_target2 = from_collection.data.insert(
        {
            "ref": Reference.to_multi_target(uuids=uuid_to2, target_collection=target),
            "name": "second",
        }
    )
    from_collection.data.insert(
        {
            "ref": Reference.to_multi_target(uuids=uuid_from_to_target1, target_collection=source),
            "name": "third",
        }
    )
    from_collection.data.insert(
        {
            "ref": Reference.to_multi_target(uuids=uuid_from_to_target2, target_collection=source),
            "name": "fourth",
        }
    )

    objects = from_collection.query.fetch_objects(
        filters=Filter(path=["ref", target, "int"]).greater_than(3)
    ).objects
    assert len(objects) == 1
    assert objects[0].properties["name"] == "second"

    objects = from_collection.query.fetch_objects(
        filters=Filter(path=["ref", source, "name"]).equal("first")
    ).objects
    assert len(objects) == 1
    assert objects[0].properties["name"] == "third"


@pytest.mark.parametrize(
    "properties,objects,where,expected_len",
    [
        (
            [
                Property(name="text", data_type=DataType.TEXT),
            ],
            [
                DataObject(properties={"text": "text"}, uuid=uuid.uuid4()),
            ],
            Filter("text").equal("text"),
            0,
        ),
        (
            [
                Property(name="text", data_type=DataType.TEXT),
            ],
            [
                DataObject(properties={"text": "there is some text in here"}, uuid=uuid.uuid4()),
            ],
            Filter("text").like("text"),
            0,
        ),
        (
            [
                Property(name="text", data_type=DataType.TEXT),
            ],
            [
                DataObject(properties={"text": "banana"}, uuid=uuid.uuid4()),
            ],
            Filter("text").like("ba*"),
            0,
        ),
        (
            [
                Property(name="texts", data_type=DataType.TEXT_ARRAY),
            ],
            [
                DataObject(properties={"texts": ["text1", "text2"]}, uuid=uuid.uuid4()),
            ],
            Filter("texts").contains_all(["text1", "text2"]),
            0,
        ),
        (
            [
                Property(name="texts", data_type=DataType.TEXT_ARRAY),
            ],
            [
                DataObject(properties={"texts": ["text1"]}, uuid=uuid.uuid4()),
                DataObject(properties={"texts": ["text2"]}, uuid=uuid.uuid4()),
            ],
            Filter("texts").contains_any(["text1"]),
            1,
        ),
        (
            [Property(name="int", data_type=DataType.INT)],
            [
                DataObject(properties={"int": 10}, uuid=uuid.uuid4()),
            ],
            Filter("int").equal(10),
            0,
        ),
        (
            [
                Property(name="int", data_type=DataType.INT),
            ],
            [
                DataObject(properties={"int": 10}, uuid=uuid.uuid4()),
            ],
            Filter("int").greater_than(5),
            0,
        ),
        (
            [
                Property(name="int", data_type=DataType.INT),
            ],
            [
                DataObject(properties={"int": 10}, uuid=uuid.uuid4()),
            ],
            Filter("int").less_than(15),
            0,
        ),
        (
            [
                Property(name="int", data_type=DataType.INT),
            ],
            [
                DataObject(properties={"int": 10}, uuid=uuid.uuid4()),
                DataObject(properties={"int": 15}, uuid=uuid.uuid4()),
            ],
            Filter("int").greater_or_equal(10),
            0,
        ),
        (
            [
                Property(name="int", data_type=DataType.INT),
            ],
            [
                DataObject(properties={"int": 10}, uuid=uuid.uuid4()),
                DataObject(properties={"int": 5}, uuid=uuid.uuid4()),
            ],
            Filter("int").less_or_equal(10),
            0,
        ),
        (
            [
                Property(name="ints", data_type=DataType.INT_ARRAY),
            ],
            [
                DataObject(properties={"ints": [1, 2]}, uuid=uuid.uuid4()),
            ],
            Filter("ints").contains_all([1, 2]),
            0,
        ),
        (
            [
                Property(name="ints", data_type=DataType.INT_ARRAY),
            ],
            [
                DataObject(properties={"ints": [1]}, uuid=uuid.uuid4()),
                DataObject(properties={"ints": [2]}, uuid=uuid.uuid4()),
            ],
            Filter("ints").contains_any([1]),
            1,
        ),
        (
            [
                Property(name="float", data_type=DataType.NUMBER),
            ],
            [
                DataObject(properties={"float": 1.0}, uuid=uuid.uuid4()),
            ],
            Filter("float").equal(1.0),
            0,
        ),
        (
            [
                Property(name="floats", data_type=DataType.NUMBER_ARRAY),
            ],
            [
                DataObject(properties={"floats": [1.0, 2.0]}, uuid=uuid.uuid4()),
            ],
            Filter("floats").contains_all([1.0, 2.0]),
            0,
        ),
        (
            [
                Property(name="floats", data_type=DataType.NUMBER_ARRAY),
            ],
            [
                DataObject(properties={"floats": [1.0]}, uuid=uuid.uuid4()),
                DataObject(properties={"floats": [2.0]}, uuid=uuid.uuid4()),
            ],
            Filter("floats").contains_any([1.0]),
            1,
        ),
        (
            [
                Property(name="float", data_type=DataType.NUMBER),
            ],
            [
                DataObject(properties={"float": 10.0}, uuid=uuid.uuid4()),
                DataObject(properties={"float": 5.0}, uuid=uuid.uuid4()),
            ],
            Filter("float").greater_than(
                5.0
            ),  # issue here, doing .greater_than(5) interprets valueInt instead of valueNumber and fails the request
            1,
        ),
        (
            [
                Property(name="bool", data_type=DataType.BOOL),
            ],
            [
                DataObject(properties={"bool": True}, uuid=uuid.uuid4()),
                DataObject(properties={"bool": False}, uuid=uuid.uuid4()),
            ],
            Filter("bool").equal(True),
            1,
        ),
        (
            [
                Property(name="bools", data_type=DataType.BOOL_ARRAY),
            ],
            [
                DataObject(properties={"bools": [True, False]}, uuid=uuid.uuid4()),
            ],
            Filter("bools").contains_all([True, False]),
            0,
        ),
        (
            [
                Property(name="bools", data_type=DataType.BOOL_ARRAY),
            ],
            [
                DataObject(properties={"bools": [True]}, uuid=uuid.uuid4()),
                DataObject(properties={"bools": [False]}, uuid=uuid.uuid4()),
            ],
            Filter("bools").contains_any([True]),
            1,
        ),
        (
            [
                Property(name="date", data_type=DataType.DATE),
            ],
            [
                DataObject(properties={"date": NOW}, uuid=uuid.uuid4()),
            ],
            Filter("date").equal(NOW),
            0,
        ),
        (
            [
                Property(name="dates", data_type=DataType.DATE_ARRAY),
            ],
            [
                DataObject(properties={"dates": [NOW, LATER]}, uuid=uuid.uuid4()),
            ],
            Filter("dates").contains_all([NOW, LATER]),
            0,
        ),
        (
            [
                Property(name="dates", data_type=DataType.DATE_ARRAY),
            ],
            [
                DataObject(properties={"dates": [NOW]}, uuid=uuid.uuid4()),
                DataObject(properties={"dates": [LATER]}, uuid=uuid.uuid4()),
            ],
            Filter("dates").contains_any([NOW]),
            1,
        ),
        (
            [
                Property(name="uuid", data_type=DataType.UUID),
            ],
            [
                DataObject(properties={"uuid": UUID1}, uuid=uuid.uuid4()),
            ],
            Filter("uuid").equal(UUID1),
            0,
        ),
        (
            [
                Property(name="uuids", data_type=DataType.UUID_ARRAY),
            ],
            [
                DataObject(properties={"uuids": [UUID1, UUID2]}, uuid=uuid.uuid4()),
            ],
            Filter("uuids").contains_all([UUID1, UUID2]),
            0,
        ),
        (
            [
                Property(name="uuids", data_type=DataType.UUID_ARRAY),
            ],
            [
                DataObject(properties={"uuids": [UUID1]}, uuid=uuid.uuid4()),
                DataObject(properties={"uuids": [UUID2]}, uuid=uuid.uuid4()),
            ],
            Filter("uuids").contains_any([UUID1]),
            1,
        ),
        (
            [
                Property(name="text", data_type=DataType.TEXT, tokenization=Tokenization.FIELD),
            ],
            [
                DataObject(properties={"text": "some name"}, vector=[1, 2, 3]),
                DataObject(properties={"text": "some other name"}, uuid=uuid.uuid4()),
            ],
            Filter("text").equal("some name"),
            1,
        ),
        (
            [
                Property(name="text", data_type=DataType.TEXT, tokenization=Tokenization.FIELD),
            ],
            [
                DataObject(properties={"text": "some name"}, vector=[1, 2, 3]),
                DataObject(properties={"text": "some other name"}, uuid=uuid.uuid4()),
            ],
            Filter("text").equal("some other name"),
            1,
        ),
        (
            [
                Property(name="text", data_type=DataType.TEXT),
                Property(name="int", data_type=DataType.INT),
            ],
            [
                DataObject(properties={"text": "Loads of money", "int": 60}, uuid=uuid.uuid4()),
                DataObject(properties={"text": "Lots of money", "int": 40}, uuid=uuid.uuid4()),
            ],
            Filter("text").equal("money"),
            0,
        ),
        (
            [
                Property(name="text", data_type=DataType.TEXT),
                Property(name="int", data_type=DataType.INT),
            ],
            [
                DataObject(properties={"int": 10}, uuid=uuid.uuid4()),
                DataObject(properties={"text": "I am ageless"}, uuid=uuid.uuid4()),
            ],
            Filter("int").is_none(True),
            1,
        ),
    ],
)
def test_delete_many_simple(
    client: weaviate.WeaviateClient,
    properties: List[Property],
    objects: List[DataObject],
    where: _FilterValue,
    expected_len: int,
):
    name = "TestDeleteManySimple"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=properties,
        inverted_index_config=Configure.inverted_index(index_null_state=True),
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many(objects)
    assert len(collection.query.fetch_objects().objects) == len(objects)

    collection.data.delete_many(where=where)
    objects = collection.query.fetch_objects().objects
    assert len(objects) == expected_len


def test_delete_many_and(client: weaviate.WeaviateClient):
    name = "TestDeleteManyAnd"
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            Property(name="Age", data_type=DataType.INT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many(
        [
            DataObject(properties={"age": 10, "name": "Timmy"}, uuid=uuid.uuid4()),
            DataObject(properties={"age": 10, "name": "Tommy"}, uuid=uuid.uuid4()),
        ]
    )
    objects = collection.query.fetch_objects().objects
    assert len(objects) == 2

    collection.data.delete_many(
        where=Filter(path="age").equal(10) & Filter(path="name").equal("Timmy")
    )

    objects = collection.query.fetch_objects().objects
    assert len(objects) == 1
    assert objects[0].properties["age"] == 10
    assert objects[0].properties["name"] == "Tommy"


def test_delete_many_or(client: weaviate.WeaviateClient):
    name = "TestDeleteManyOr"
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            Property(name="Age", data_type=DataType.INT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many(
        [
            DataObject(properties={"age": 10, "name": "Timmy"}, uuid=uuid.uuid4()),
            DataObject(properties={"age": 20, "name": "Tim"}, uuid=uuid.uuid4()),
            DataObject(properties={"age": 30, "name": "Timothy"}, uuid=uuid.uuid4()),
        ]
    )
    objects = collection.query.fetch_objects().objects
    assert len(objects) == 3

    collection.data.delete_many(where=Filter(path="age").equal(10) | Filter(path="age").equal(30))
    objects = collection.query.fetch_objects().objects
    assert len(objects) == 1
    assert objects[0].properties["age"] == 20
    assert objects[0].properties["name"] == "Tim"


def test_delete_many_return(client: weaviate.WeaviateClient):
    name = "TestDeleteManyReturn"
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many(
        [
            DataObject(properties={"name": "delet me"}, uuid=uuid.uuid4()),
        ]
    )
    ret = collection.data.delete_many(where=Filter(path="name").equal("delet me"))
    assert ret.failed == 0
    assert ret.matches == 1
    assert ret.objects is None
    assert ret.successful == 1



from typing import Generic, List, Optional, Type, overload

from weaviate.collections.classes.filters import (
    _Filters,
)
from weaviate.collections.classes.grpc import (
    MetadataQuery,
    PROPERTIES,
)
from weaviate.collections.classes.internal import (
    _Generative,
    _GenerativeReturn,
    _GroupBy,
    _GroupByReturn,
    _QueryReturn,
    QueryReturn,
    GenerativeReturn,
    GroupByReturn,
    ReturnProperties,
)
from weaviate.collections.classes.types import Properties, TProperties
from weaviate.collections.queries.base import _Grpc
from weaviate.types import UUID


class _NearObjectQuery(Generic[Properties], _Grpc[Properties]):
    @overload
    def near_object(
        self,
        near_object: UUID,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> _QueryReturn[Properties]:
        ...

    @overload
    def near_object(
        self,
        near_object: UUID,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        *,
        return_properties: Type[TProperties],
    ) -> _QueryReturn[TProperties]:
        ...

    def near_object(
        self,
        near_object: UUID,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
    ) -> QueryReturn[Properties, TProperties]:
        """Search for objects in this collection by another object using a vector-based similarity search.

        See the [docs](https://weaviate.io/developers/weaviate/api/graphql/search-operators#nearobject) for a more detailed explanation.

        Arguments:
            `near_object`
                The UUID of the object to search on, REQUIRED.
            `certainty`
                The minimum similarity score to return. If not specified, the default certainty specified by the server is used.
            `distance`
                The maximum distance to search. If not specified, the default distance specified by the server is used.
            `limit`
                The maximum number of results to return. If not specified, the default limit specified by the server is returned.
            `auto_limit`
                The maximum number of [autocut](https://weaviate.io/developers/weaviate/api/graphql/additional-operators#autocut) results to return. If not specified, no limit is applied.
            `filters`
                The filters to apply to the search.
            `return_metadata`
                The metadata to return for each object.
            `return_properties`
                The properties to return for each object.

        Returns:
            A `_QueryReturn` object that includes the searched objects.

        Raises:
            `weaviate.exceptions.WeaviateGrpcError`:
                If the request to the Weaviate server fails.
        """
        ret_properties, ret_metadata = self._parse_return_properties(return_properties)
        res = self._query().near_object(
            near_object=near_object,
            certainty=certainty,
            distance=distance,
            limit=limit,
            autocut=auto_limit,
            filters=filters,
            return_metadata=return_metadata,
            return_properties=ret_properties,
        )
        return self._result_to_query_return(res, return_properties)


class _NearObjectGenerate(Generic[Properties], _Grpc[Properties]):
    @overload
    def near_object(
        self,
        near_object: UUID,
        single_prompt: Optional[str] = None,
        grouped_task: Optional[str] = None,
        grouped_properties: Optional[List[str]] = None,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> _GenerativeReturn[Properties]:
        ...

    @overload
    def near_object(
        self,
        near_object: UUID,
        single_prompt: Optional[str] = None,
        grouped_task: Optional[str] = None,
        grouped_properties: Optional[List[str]] = None,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        *,
        return_properties: Type[TProperties],
    ) -> _GenerativeReturn[TProperties]:
        ...

    def near_object(
        self,
        near_object: UUID,
        single_prompt: Optional[str] = None,
        grouped_task: Optional[str] = None,
        grouped_properties: Optional[List[str]] = None,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
    ) -> GenerativeReturn[Properties, TProperties]:
        """Perform retrieval-augmented generation (RaG) on the results of a by-object object search in this collection using a vector-based similarity search.

        See the [docs](https://weaviate.io/developers/weaviate/api/graphql/search-operators#nearobject) for a more detailed explanation.

        Arguments:
            `near_object`
                The UUID of the object to search on, REQUIRED.
            `certainty`
                The minimum similarity score to return. If not specified, the default certainty specified by the server is used.
            `distance`
                The maximum distance to search. If not specified, the default distance specified by the server is used.
            `limit`
                The maximum number of results to return. If not specified, the default limit specified by the server is returned.
            `auto_limit`
                The maximum number of [autocut](https://weaviate.io/developers/weaviate/api/graphql/additional-operators#autocut) results to return. If not specified, no limit is applied.
            `filters`
                The filters to apply to the search.
            `return_metadata`
                The metadata to return for each object.
            `return_properties`
                The properties to return for each object.

        Returns:
            A `_GenerativeReturn` object that includes the searched objects with per-object generated results and group generated results.

        Raises:
            `weaviate.exceptions.WeaviateGrpcError`:
                If the request to the Weaviate server fails.
        """
        ret_properties, ret_metadata = self._parse_return_properties(return_properties)
        res = self._query().near_object(
            near_object=near_object,
            certainty=certainty,
            distance=distance,
            limit=limit,
            autocut=auto_limit,
            filters=filters,
            generative=_Generative(
                single=single_prompt,
                grouped=grouped_task,
                grouped_properties=grouped_properties,
            ),
            return_metadata=return_metadata or ret_metadata,
            return_properties=ret_properties,
        )
        return self._result_to_generative_return(res, return_properties)


class _NearObjectGroupBy(Generic[Properties], _Grpc[Properties]):
    @overload
    def near_object(
        self,
        near_object: UUID,
        group_by_property: str,
        number_of_groups: int,
        objects_per_group: int,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> _GroupByReturn[Properties]:
        ...

    @overload
    def near_object(
        self,
        near_object: UUID,
        group_by_property: str,
        number_of_groups: int,
        objects_per_group: int,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        *,
        return_properties: Type[TProperties],
    ) -> _GroupByReturn[TProperties]:
        ...

    def near_object(
        self,
        near_object: UUID,
        group_by_property: str,
        number_of_groups: int,
        objects_per_group: int,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
    ) -> GroupByReturn[Properties, TProperties]:
        """Group the results of a by-object object search in this collection using a vector-based similarity search.

        See the [docs](https://weaviate.io/developers/weaviate/api/graphql/search-operators#nearobject) for a more detailed explanation.

        Arguments:
            `near_object`
                The UUID of the object to search on, REQUIRED.
            `group_by_property`
                The property to group by, REQUIRED.
            `number_of_groups`
                The number of groups to return, REQUIRED.
            `objects_per_group`
                The number of objects to return per group, REQUIRED.
            `certainty`
                The minimum similarity score to return. If not specified, the default certainty specified by the server is used.
            `distance`
                The maximum distance to search. If not specified, the default distance specified by the server is used.
            `limit`
                The maximum number of results to return. If not specified, the default limit specified by the server is returned.
            `auto_limit`
                The maximum number of [autocut](https://weaviate.io/developers/weaviate/api/graphql/additional-operators#autocut) results to return. If not specified, no limit is applied.
            `filters`
                The filters to apply to the search.
            `return_metadata`
                The metadata to return for each object.
            `return_properties`
                The properties to return for each object.

         Returns:
            A `_GroupByReturn` object that includes the searched objects grouped by the specified property.

        Raises:
            `weaviate.exceptions.WeaviateGrpcError`:
                If the request to the Weaviate server fails.
        """
        ret_properties, ret_metadata = self._parse_return_properties(return_properties)
        res = self._query().near_object(
            near_object=near_object,
            certainty=certainty,
            distance=distance,
            limit=limit,
            autocut=auto_limit,
            filters=filters,
            group_by=_GroupBy(
                prop=group_by_property,
                number_of_groups=number_of_groups,
                objects_per_group=objects_per_group,
            ),
            return_metadata=return_metadata or ret_metadata,
            return_properties=ret_properties,
        )
        return self._result_to_groupby_return(res, return_properties)



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Union

from weaviate.types import UUID
from weaviate.proto.v1 import search_get_pb2


class _Operator(str, Enum):
    EQUAL = "Equal"
    NOT_EQUAL = "NotEqual"
    LESS_THAN = "LessThan"
    LESS_THAN_EQUAL = "LessThanEqual"
    GREATER_THAN = "GreaterThan"
    GREATER_THAN_EQUAL = "GreaterThanEqual"
    LIKE = "Like"
    IS_NULL = "IsNull"
    CONTAINS_ANY = "ContainsAny"
    CONTAINS_ALL = "ContainsAll"
    AND = "And"
    OR = "Or"

    def _to_grpc(self) -> search_get_pb2.Filters.Operator:
        if self == _Operator.EQUAL:
            return search_get_pb2.Filters.OPERATOR_EQUAL
        elif self == _Operator.NOT_EQUAL:
            return search_get_pb2.Filters.OPERATOR_NOT_EQUAL
        elif self == _Operator.LESS_THAN:
            return search_get_pb2.Filters.OPERATOR_LESS_THAN
        elif self == _Operator.LESS_THAN_EQUAL:
            return search_get_pb2.Filters.OPERATOR_LESS_THAN_EQUAL
        elif self == _Operator.GREATER_THAN:
            return search_get_pb2.Filters.OPERATOR_GREATER_THAN
        elif self == _Operator.GREATER_THAN_EQUAL:
            return search_get_pb2.Filters.OPERATOR_GREATER_THAN_EQUAL
        elif self == _Operator.LIKE:
            return search_get_pb2.Filters.OPERATOR_LIKE
        elif self == _Operator.IS_NULL:
            return search_get_pb2.Filters.OPERATOR_IS_NULL
        elif self == _Operator.CONTAINS_ANY:
            return search_get_pb2.Filters.OPERATOR_CONTAINS_ANY
        elif self == _Operator.CONTAINS_ALL:
            return search_get_pb2.Filters.OPERATOR_CONTAINS_ALL
        elif self == _Operator.AND:
            return search_get_pb2.Filters.OPERATOR_AND
        elif self == _Operator.OR:
            return search_get_pb2.Filters.OPERATOR_OR
        else:
            raise ValueError(f"Unknown operator {self}")


class _Filters:
    def __and__(self, other: "_Filters") -> "_FilterAnd":
        return _FilterAnd(self, other)

    def __or__(self, other: "_Filters") -> "_FilterOr":
        return _FilterOr(self, other)


class _FilterAnd(_Filters):
    def __init__(self, *args: _Filters):
        self.filters: List[_Filters] = list(args)

    # replace with the following once 3.11 is the minimum version
    #     Operator: weaviate_pb2.Filters.OperatorType = weaviate_pb2.Filters.OperatorAnd
    @property
    def operator(self) -> _Operator:
        return _Operator.AND


class _FilterOr(_Filters):
    def __init__(self, *args: _Filters):
        self.filters: List[_Filters] = list(args)

    # replace with the following once 3.11 is the minimum version
    #     Operator: weaviate_pb2.Filters.OperatorType = weaviate_pb2.Filters.OperatorOr
    @property
    def operator(self) -> _Operator:
        return _Operator.OR


FilterValuesList = Union[List[str], List[bool], List[int], List[float], List[datetime], List[UUID]]
FilterValues = Union[int, float, str, bool, datetime, UUID, None, FilterValuesList]


@dataclass
class _FilterValue(_Filters):
    path: Union[str, List[str]]
    value: FilterValues
    operator: _Operator


class Filter:
    """Define a filter based on a property to be used when querying a collection.

    Use the `__init__` method to define the path to the property to be filtered on and then
    use the methods of this class to define the condition to filter on.

    To combine multiple filters, you can use `&` or `|` operators for each `AND` or `OR` operations, e.g.,
        `Filter("name").equal("John") & Filter("age").greater_than(18)`

    See [the docs](https://weaviate.io/developers/weaviate/search/filters) for more details!
    """

    def __init__(self, path: Union[str, List[str]], length: bool = False):
        """Initialise the filter.

        Arguments:
            `path`
                The path to the property to be filtered on.
                    This can be a single string in the case of a top-level property or a list of strings in the case of a nested cross-ref property.
            `length`
                If `True`, the length of the property will be used in the filter. Defaults to `False`.
                    This is only valid for properties of type `string` or `text`. The inverted index must also be configured to index the property length.
        """
        if isinstance(path, str):
            path = [path]
        if length:
            path[-1] = "len(" + path[-1] + ")"
        self.__internal_path = path

    def is_none(self, val: bool) -> _FilterValue:
        """Filter on whether the property is `None`."""
        return _FilterValue(path=self.__internal_path, value=val, operator=_Operator.IS_NULL)

    def contains_any(self, val: FilterValuesList) -> _FilterValue:
        """Filter on whether the property contains any of the given values."""
        return _FilterValue(
            path=self.__internal_path,
            value=val,
            operator=_Operator.CONTAINS_ANY,
        )

    def contains_all(self, val: FilterValuesList) -> _FilterValue:
        """Filter on whether the property contains all of the given values."""
        return _FilterValue(
            path=self.__internal_path,
            value=val,
            operator=_Operator.CONTAINS_ALL,
        )

    def equal(self, val: FilterValues) -> _FilterValue:
        """Filter on whether the property is equal to the given value."""
        return _FilterValue(path=self.__internal_path, value=val, operator=_Operator.EQUAL)

    def not_equal(self, val: FilterValues) -> _FilterValue:
        """Filter on whether the property is not equal to the given value."""
        return _FilterValue(path=self.__internal_path, value=val, operator=_Operator.NOT_EQUAL)

    def less_than(self, val: FilterValues) -> _FilterValue:
        """Filter on whether the property is less than the given value."""
        return _FilterValue(path=self.__internal_path, value=val, operator=_Operator.LESS_THAN)

    def less_or_equal(self, val: FilterValues) -> _FilterValue:
        """Filter on whether the property is less than or equal to the given value."""
        return _FilterValue(
            path=self.__internal_path,
            value=val,
            operator=_Operator.LESS_THAN_EQUAL,
        )

    def greater_than(self, val: FilterValues) -> _FilterValue:
        """Filter on whether the property is greater than the given value."""
        return _FilterValue(
            path=self.__internal_path,
            value=val,
            operator=_Operator.GREATER_THAN,
        )

    def greater_or_equal(self, val: FilterValues) -> _FilterValue:
        """Filter on whether the property is greater than or equal to the given value."""
        return _FilterValue(
            path=self.__internal_path,
            value=val,
            operator=_Operator.GREATER_THAN_EQUAL,
        )

    def like(self, val: str) -> _FilterValue:
        """Filter on whether the property is like the given value.

        This filter can make use of `*` and `?` as wildcards. See [the docs](https://weaviate.io/developers/weaviate/search/filters#by-partial-matches-text) for more details.
        """
        return _FilterValue(path=self.__internal_path, value=val, operator=_Operator.LIKE)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/search_get.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from weaviate.proto.v1 import base_pb2 as v1_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13v1/search_get.proto\x12\x0bweaviate.v1\x1a\x1cgoogle/protobuf/struct.proto\x1a\rv1/base.proto"\xcb\x08\n\rSearchRequest\x12\x12\n\ncollection\x18\x01 \x01(\t\x12\x0e\n\x06tenant\x18\n \x01(\t\x12=\n\x11\x63onsistency_level\x18\x0b \x01(\x0e\x32\x1d.weaviate.v1.ConsistencyLevelH\x00\x88\x01\x01\x12\x37\n\nproperties\x18\x14 \x01(\x0b\x32\x1e.weaviate.v1.PropertiesRequestH\x01\x88\x01\x01\x12\x33\n\x08metadata\x18\x15 \x01(\x0b\x32\x1c.weaviate.v1.MetadataRequestH\x02\x88\x01\x01\x12+\n\x08group_by\x18\x16 \x01(\x0b\x32\x14.weaviate.v1.GroupByH\x03\x88\x01\x01\x12\r\n\x05limit\x18\x1e \x01(\r\x12\x0e\n\x06offset\x18\x1f \x01(\r\x12\x0f\n\x07\x61utocut\x18  \x01(\r\x12\r\n\x05\x61\x66ter\x18! \x01(\t\x12$\n\x07sort_by\x18" \x03(\x0b\x32\x13.weaviate.v1.SortBy\x12*\n\x07\x66ilters\x18( \x01(\x0b\x32\x14.weaviate.v1.FiltersH\x04\x88\x01\x01\x12/\n\rhybrid_search\x18) \x01(\x0b\x32\x13.weaviate.v1.HybridH\x05\x88\x01\x01\x12+\n\x0b\x62m25_search\x18* \x01(\x0b\x32\x11.weaviate.v1.BM25H\x06\x88\x01\x01\x12\x31\n\x0bnear_vector\x18+ \x01(\x0b\x32\x17.weaviate.v1.NearVectorH\x07\x88\x01\x01\x12\x31\n\x0bnear_object\x18, \x01(\x0b\x32\x17.weaviate.v1.NearObjectH\x08\x88\x01\x01\x12\x33\n\tnear_text\x18- \x01(\x0b\x32\x1b.weaviate.v1.NearTextSearchH\t\x88\x01\x01\x12\x35\n\nnear_image\x18. \x01(\x0b\x32\x1c.weaviate.v1.NearImageSearchH\n\x88\x01\x01\x12\x35\n\nnear_audio\x18/ \x01(\x0b\x32\x1c.weaviate.v1.NearAudioSearchH\x0b\x88\x01\x01\x12\x35\n\nnear_video\x18\x30 \x01(\x0b\x32\x1c.weaviate.v1.NearVideoSearchH\x0c\x88\x01\x01\x12\x36\n\ngenerative\x18< \x01(\x0b\x32\x1d.weaviate.v1.GenerativeSearchH\r\x88\x01\x01\x42\x14\n\x12_consistency_levelB\r\n\x0b_propertiesB\x0b\n\t_metadataB\x0b\n\t_group_byB\n\n\x08_filtersB\x10\n\x0e_hybrid_searchB\x0e\n\x0c_bm25_searchB\x0e\n\x0c_near_vectorB\x0e\n\x0c_near_objectB\x0c\n\n_near_textB\r\n\x0b_near_imageB\r\n\x0b_near_audioB\r\n\x0b_near_videoB\r\n\x0b_generative"L\n\x07GroupBy\x12\x0c\n\x04path\x18\x01 \x03(\t\x12\x18\n\x10number_of_groups\x18\x02 \x01(\x05\x12\x19\n\x11objects_per_group\x18\x03 \x01(\x05")\n\x06SortBy\x12\x11\n\tascending\x18\x01 \x01(\x08\x12\x0c\n\x04path\x18\x02 \x03(\t"m\n\x10GenerativeSearch\x12\x1e\n\x16single_response_prompt\x18\x01 \x01(\t\x12\x1d\n\x15grouped_response_task\x18\x02 \x01(\t\x12\x1a\n\x12grouped_properties\x18\x03 \x03(\t"\x1b\n\tTextArray\x12\x0e\n\x06values\x18\x01 \x03(\t"\x1a\n\x08IntArray\x12\x0e\n\x06values\x18\x01 \x03(\x03"\x1d\n\x0bNumberArray\x12\x0e\n\x06values\x18\x01 \x03(\x01"\x1e\n\x0c\x42ooleanArray\x12\x0e\n\x06values\x18\x01 \x03(\x08"\x95\x06\n\x07\x46ilters\x12/\n\x08operator\x18\x01 \x01(\x0e\x32\x1d.weaviate.v1.Filters.Operator\x12\n\n\x02on\x18\x02 \x03(\t\x12%\n\x07\x66ilters\x18\x03 \x03(\x0b\x32\x14.weaviate.v1.Filters\x12\x14\n\nvalue_text\x18\x04 \x01(\tH\x00\x12\x13\n\tvalue_int\x18\x05 \x01(\x03H\x00\x12\x17\n\rvalue_boolean\x18\x06 \x01(\x08H\x00\x12\x16\n\x0cvalue_number\x18\x07 \x01(\x02H\x00\x12\x32\n\x10value_text_array\x18\t \x01(\x0b\x32\x16.weaviate.v1.TextArrayH\x00\x12\x30\n\x0fvalue_int_array\x18\n \x01(\x0b\x32\x15.weaviate.v1.IntArrayH\x00\x12\x38\n\x13value_boolean_array\x18\x0b \x01(\x0b\x32\x19.weaviate.v1.BooleanArrayH\x00\x12\x36\n\x12value_number_array\x18\x0c \x01(\x0b\x32\x18.weaviate.v1.NumberArrayH\x00"\xe3\x02\n\x08Operator\x12\x18\n\x14OPERATOR_UNSPECIFIED\x10\x00\x12\x12\n\x0eOPERATOR_EQUAL\x10\x01\x12\x16\n\x12OPERATOR_NOT_EQUAL\x10\x02\x12\x19\n\x15OPERATOR_GREATER_THAN\x10\x03\x12\x1f\n\x1bOPERATOR_GREATER_THAN_EQUAL\x10\x04\x12\x16\n\x12OPERATOR_LESS_THAN\x10\x05\x12\x1c\n\x18OPERATOR_LESS_THAN_EQUAL\x10\x06\x12\x10\n\x0cOPERATOR_AND\x10\x07\x12\x0f\n\x0bOPERATOR_OR\x10\x08\x12\x1d\n\x19OPERATOR_WITHIN_GEO_RANGE\x10\t\x12\x11\n\rOPERATOR_LIKE\x10\n\x12\x14\n\x10OPERATOR_IS_NULL\x10\x0b\x12\x19\n\x15OPERATOR_CONTAINS_ANY\x10\x0c\x12\x19\n\x15OPERATOR_CONTAINS_ALL\x10\rB\x0c\n\ntest_value"\xcc\x01\n\x0fMetadataRequest\x12\x0c\n\x04uuid\x18\x01 \x01(\x08\x12\x0e\n\x06vector\x18\x02 \x01(\x08\x12\x1a\n\x12\x63reation_time_unix\x18\x03 \x01(\x08\x12\x1d\n\x15last_update_time_unix\x18\x04 \x01(\x08\x12\x10\n\x08\x64istance\x18\x05 \x01(\x08\x12\x11\n\tcertainty\x18\x06 \x01(\x08\x12\r\n\x05score\x18\x07 \x01(\x08\x12\x15\n\rexplain_score\x18\x08 \x01(\x08\x12\x15\n\ris_consistent\x18\t \x01(\x08"\xab\x01\n\x11PropertiesRequest\x12\x1a\n\x12non_ref_properties\x18\x01 \x03(\t\x12\x39\n\x0eref_properties\x18\x02 \x03(\x0b\x32!.weaviate.v1.RefPropertiesRequest\x12?\n\x11object_properties\x18\x03 \x03(\x0b\x32$.weaviate.v1.ObjectPropertiesRequest"\x8b\x01\n\x17ObjectPropertiesRequest\x12\x11\n\tprop_name\x18\x01 \x01(\t\x12\x1c\n\x14primitive_properties\x18\x02 \x03(\t\x12?\n\x11object_properties\x18\x03 \x03(\x0b\x32$.weaviate.v1.ObjectPropertiesRequest"\xe2\x01\n\x06Hybrid\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\x12\x0e\n\x06vector\x18\x03 \x03(\x02\x12\r\n\x05\x61lpha\x18\x04 \x01(\x02\x12\x33\n\x0b\x66usion_type\x18\x05 \x01(\x0e\x32\x1e.weaviate.v1.Hybrid.FusionType"a\n\nFusionType\x12\x1b\n\x17\x46USION_TYPE_UNSPECIFIED\x10\x00\x12\x16\n\x12\x46USION_TYPE_RANKED\x10\x01\x12\x1e\n\x1a\x46USION_TYPE_RELATIVE_SCORE\x10\x02"\xad\x02\n\x0eNearTextSearch\x12\r\n\x05query\x18\x01 \x03(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x36\n\x07move_to\x18\x04 \x01(\x0b\x32 .weaviate.v1.NearTextSearch.MoveH\x02\x88\x01\x01\x12\x38\n\tmove_away\x18\x05 \x01(\x0b\x32 .weaviate.v1.NearTextSearch.MoveH\x03\x88\x01\x01\x1a\x36\n\x04Move\x12\r\n\x05\x66orce\x18\x01 \x01(\x02\x12\x10\n\x08\x63oncepts\x18\x02 \x03(\t\x12\r\n\x05uuids\x18\x03 \x03(\tB\x0c\n\n_certaintyB\x0b\n\t_distanceB\n\n\x08_move_toB\x0c\n\n_move_away"j\n\x0fNearImageSearch\x12\r\n\x05image\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance"j\n\x0fNearAudioSearch\x12\r\n\x05\x61udio\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance"j\n\x0fNearVideoSearch\x12\r\n\x05video\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance")\n\x04\x42M25\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t"\xb1\x01\n\x14RefPropertiesRequest\x12\x1a\n\x12reference_property\x18\x01 \x01(\t\x12\x32\n\nproperties\x18\x02 \x01(\x0b\x32\x1e.weaviate.v1.PropertiesRequest\x12.\n\x08metadata\x18\x03 \x01(\x0b\x32\x1c.weaviate.v1.MetadataRequest\x12\x19\n\x11target_collection\x18\x04 \x01(\t"f\n\nNearVector\x12\x0e\n\x06vector\x18\x01 \x03(\x02\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance"b\n\nNearObject\x12\n\n\x02id\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance"\xc3\x01\n\x0bSearchReply\x12\x0c\n\x04took\x18\x01 \x01(\x02\x12*\n\x07results\x18\x02 \x03(\x0b\x32\x19.weaviate.v1.SearchResult\x12&\n\x19generative_grouped_result\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x34\n\x10group_by_results\x18\x04 \x03(\x0b\x32\x1a.weaviate.v1.GroupByResultB\x1c\n\x1a_generative_grouped_result"\x90\x01\n\rGroupByResult\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cmin_distance\x18\x02 \x01(\x02\x12\x14\n\x0cmax_distance\x18\x03 \x01(\x02\x12\x19\n\x11number_of_objects\x18\x04 \x01(\x03\x12*\n\x07objects\x18\x05 \x03(\x0b\x32\x19.weaviate.v1.SearchResult"p\n\x0cSearchResult\x12\x31\n\nproperties\x18\x01 \x01(\x0b\x32\x1d.weaviate.v1.PropertiesResult\x12-\n\x08metadata\x18\x02 \x01(\x0b\x32\x1b.weaviate.v1.MetadataResult"\xc6\x03\n\x0eMetadataResult\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06vector\x18\x02 \x03(\x02\x12\x1a\n\x12\x63reation_time_unix\x18\x03 \x01(\x03\x12"\n\x1a\x63reation_time_unix_present\x18\x04 \x01(\x08\x12\x1d\n\x15last_update_time_unix\x18\x05 \x01(\x03\x12%\n\x1dlast_update_time_unix_present\x18\x06 \x01(\x08\x12\x10\n\x08\x64istance\x18\x07 \x01(\x02\x12\x18\n\x10\x64istance_present\x18\x08 \x01(\x08\x12\x11\n\tcertainty\x18\t \x01(\x02\x12\x19\n\x11\x63\x65rtainty_present\x18\n \x01(\x08\x12\r\n\x05score\x18\x0b \x01(\x02\x12\x15\n\rscore_present\x18\x0c \x01(\x08\x12\x15\n\rexplain_score\x18\r \x01(\t\x12\x1d\n\x15\x65xplain_score_present\x18\x0e \x01(\x08\x12\x1a\n\ris_consistent\x18\x0f \x01(\x08H\x00\x88\x01\x01\x12\x12\n\ngenerative\x18\x10 \x01(\t\x12\x1a\n\x12generative_present\x18\x11 \x01(\x08\x42\x10\n\x0e_is_consistent"\xd1\x04\n\x10PropertiesResult\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x33\n\tref_props\x18\x02 \x03(\x0b\x32 .weaviate.v1.RefPropertiesResult\x12\x19\n\x11target_collection\x18\x03 \x01(\t\x12-\n\x08metadata\x18\x04 \x01(\x0b\x32\x1b.weaviate.v1.MetadataResult\x12\x43\n\x17number_array_properties\x18\x05 \x03(\x0b\x32".weaviate.v1.NumberArrayProperties\x12=\n\x14int_array_properties\x18\x06 \x03(\x0b\x32\x1f.weaviate.v1.IntArrayProperties\x12?\n\x15text_array_properties\x18\x07 \x03(\x0b\x32 .weaviate.v1.TextArrayProperties\x12\x45\n\x18\x62oolean_array_properties\x18\x08 \x03(\x0b\x32#.weaviate.v1.BooleanArrayProperties\x12\x38\n\x11object_properties\x18\t \x03(\x0b\x32\x1d.weaviate.v1.ObjectProperties\x12\x43\n\x17object_array_properties\x18\n \x03(\x0b\x32".weaviate.v1.ObjectArrayProperties"[\n\x13RefPropertiesResult\x12\x31\n\nproperties\x18\x01 \x03(\x0b\x32\x1d.weaviate.v1.PropertiesResult\x12\x11\n\tprop_name\x18\x02 \x01(\tBs\n#io.weaviate.client.grpc.protocol.v1B\x16WeaviateProtoSearchGetZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "v1.search_get_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n#io.weaviate.client.grpc.protocol.v1B\026WeaviateProtoSearchGetZ4github.com/weaviate/weaviate/grpc/generated;protocol"
    _globals["_SEARCHREQUEST"]._serialized_start = 82
    _globals["_SEARCHREQUEST"]._serialized_end = 1181
    _globals["_GROUPBY"]._serialized_start = 1183
    _globals["_GROUPBY"]._serialized_end = 1259
    _globals["_SORTBY"]._serialized_start = 1261
    _globals["_SORTBY"]._serialized_end = 1302
    _globals["_GENERATIVESEARCH"]._serialized_start = 1304
    _globals["_GENERATIVESEARCH"]._serialized_end = 1413
    _globals["_TEXTARRAY"]._serialized_start = 1415
    _globals["_TEXTARRAY"]._serialized_end = 1442
    _globals["_INTARRAY"]._serialized_start = 1444
    _globals["_INTARRAY"]._serialized_end = 1470
    _globals["_NUMBERARRAY"]._serialized_start = 1472
    _globals["_NUMBERARRAY"]._serialized_end = 1501
    _globals["_BOOLEANARRAY"]._serialized_start = 1503
    _globals["_BOOLEANARRAY"]._serialized_end = 1533
    _globals["_FILTERS"]._serialized_start = 1536
    _globals["_FILTERS"]._serialized_end = 2325
    _globals["_FILTERS_OPERATOR"]._serialized_start = 1956
    _globals["_FILTERS_OPERATOR"]._serialized_end = 2311
    _globals["_METADATAREQUEST"]._serialized_start = 2328
    _globals["_METADATAREQUEST"]._serialized_end = 2532
    _globals["_PROPERTIESREQUEST"]._serialized_start = 2535
    _globals["_PROPERTIESREQUEST"]._serialized_end = 2706
    _globals["_OBJECTPROPERTIESREQUEST"]._serialized_start = 2709
    _globals["_OBJECTPROPERTIESREQUEST"]._serialized_end = 2848
    _globals["_HYBRID"]._serialized_start = 2851
    _globals["_HYBRID"]._serialized_end = 3077
    _globals["_HYBRID_FUSIONTYPE"]._serialized_start = 2980
    _globals["_HYBRID_FUSIONTYPE"]._serialized_end = 3077
    _globals["_NEARTEXTSEARCH"]._serialized_start = 3080
    _globals["_NEARTEXTSEARCH"]._serialized_end = 3381
    _globals["_NEARTEXTSEARCH_MOVE"]._serialized_start = 3274
    _globals["_NEARTEXTSEARCH_MOVE"]._serialized_end = 3328
    _globals["_NEARIMAGESEARCH"]._serialized_start = 3383
    _globals["_NEARIMAGESEARCH"]._serialized_end = 3489
    _globals["_NEARAUDIOSEARCH"]._serialized_start = 3491
    _globals["_NEARAUDIOSEARCH"]._serialized_end = 3597
    _globals["_NEARVIDEOSEARCH"]._serialized_start = 3599
    _globals["_NEARVIDEOSEARCH"]._serialized_end = 3705
    _globals["_BM25"]._serialized_start = 3707
    _globals["_BM25"]._serialized_end = 3748
    _globals["_REFPROPERTIESREQUEST"]._serialized_start = 3751
    _globals["_REFPROPERTIESREQUEST"]._serialized_end = 3928
    _globals["_NEARVECTOR"]._serialized_start = 3930
    _globals["_NEARVECTOR"]._serialized_end = 4032
    _globals["_NEAROBJECT"]._serialized_start = 4034
    _globals["_NEAROBJECT"]._serialized_end = 4132
    _globals["_SEARCHREPLY"]._serialized_start = 4135
    _globals["_SEARCHREPLY"]._serialized_end = 4330
    _globals["_GROUPBYRESULT"]._serialized_start = 4333
    _globals["_GROUPBYRESULT"]._serialized_end = 4477
    _globals["_SEARCHRESULT"]._serialized_start = 4479
    _globals["_SEARCHRESULT"]._serialized_end = 4591
    _globals["_METADATARESULT"]._serialized_start = 4594
    _globals["_METADATARESULT"]._serialized_end = 5048
    _globals["_PROPERTIESRESULT"]._serialized_start = 5051
    _globals["_PROPERTIESRESULT"]._serialized_end = 5644
    _globals["_REFPROPERTIESRESULT"]._serialized_start = 5646
    _globals["_REFPROPERTIESRESULT"]._serialized_end = 5737
# @@protoc_insertion_point(module_scope)
